#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Testing functions for the command line interface.
"""

import os

import pyben
import pytest

from tests import file1, file2, filemeta2, torrents
from torrentfile.interactive import select_action
from torrentfile.utils import normalize_piece_length

MOCK = "torrentfile.interactive.get_input"


def test_fixtures():
    """
    Test the fixtures used in module.
    """
    assert filemeta2 and file1 and file2


def test_interactive_create(monkeypatch, file1):
    """
    Test creating torrent interactively.
    """
    mapping = [
        "create",
        "",
        "",
        "",
        "",
        "",
        "",
        "",
        file1,
        str(file1) + ".torrent",
        "",
    ]
    it = iter(mapping)
    monkeypatch.setattr(MOCK, lambda *_: next(it))
    select_action()
    assert os.path.exists(str(file1) + ".torrent")


@pytest.mark.parametrize("version", ["1", "2", "3"])
@pytest.mark.parametrize("piece_length", ["23", "18", "131072"])
@pytest.mark.parametrize("announce", ["url1", "urla urlb urlc"])
@pytest.mark.parametrize("url_list", ["ftp url2", "ftp1 ftp2 ftp3"])
@pytest.mark.parametrize("comment", ["Some Comment", "No Comment"])
@pytest.mark.parametrize("source", ["Do", "Ra", "Me"])
def test_inter_create_full(
    file1,
    piece_length,
    announce,
    comment,
    source,
    url_list,
    version,
    monkeypatch,
):
    """
    Test creating torrent interactively with many parameters.
    """
    mapping = [
        "create",
        piece_length,
        announce,
        url_list,
        url_list,
        comment,
        source,
        "Y",
        file1,
        str(file1) + ".torrent",
        version,
    ]
    it = iter(mapping)
    monkeypatch.setattr(MOCK, lambda *_: next(it))
    select_action()
    meta = pyben.load(str(file1) + ".torrent")
    assert meta["info"]["source"] == source
    assert meta["info"]["piece length"] == normalize_piece_length(piece_length)
    assert meta["info"]["comment"] == comment
    assert meta["url-list"] == url_list.split()


@pytest.mark.parametrize("announce", ["url1"])
@pytest.mark.parametrize("url_list", ["ftp url2", "ftp1 ftp2 ftp3"])
@pytest.mark.parametrize("comment", ["Some Comment", "No Comment"])
@pytest.mark.parametrize("source", ["Fa", "So", "La"])
def test_inter_edit_full(filemeta2, announce, comment, source, url_list,
                         monkeypatch):
    """
    Test editing torrent file interactively.
    """
    seq = [
        "edit",
        filemeta2,
        "4",
        announce,
        "1",
        comment,
        "2",
        source,
        "5",
        url_list,
        "",
        "6",
        "Y",
        "DONE",
    ]
    it = iter(seq)
    monkeypatch.setattr(MOCK, lambda *_: next(it))
    select_action()
    meta1 = pyben.load(filemeta2)
    assert meta1["info"]["source"] == source
    assert meta1["info"]["comment"] == comment
    assert meta1["url-list"] == url_list.split()
    assert meta1["info"]["private"] == 1


@pytest.mark.parametrize("announce", ["urla urlb urlc", "urld url2"])
@pytest.mark.parametrize("urllist", ["ftp url2", "ftp1 ftp2 ftp3"])
@pytest.mark.parametrize("cmnt", ["Some Comment"])
@pytest.mark.parametrize("srce", ["Do", "Ra"])
def test_inter_edit_cli(filemeta2, announce, cmnt, srce, urllist, monkeypatch):
    """
    Test editing torrent interactively from CLI.
    """
    seq = [
        "edit",
        filemeta2,
        "4",
        announce,
        "1",
        cmnt,
        "2",
        srce,
        "5",
        urllist,
        urllist,
        "6",
        "Y",
        "DONE",
    ]
    it = iter(seq)
    monkeypatch.setattr(MOCK, lambda *_: next(it))
    select_action()
    meta2 = pyben.load(filemeta2)
    assert meta2["info"]["source"] == srce
    assert meta2["info"]["comment"] == cmnt
    assert meta2["url-list"] == urllist.split()
    assert meta2["info"]["private"] == 1


@pytest.mark.parametrize("torrentclass", torrents())
def test_inter_recheck(torrentclass, monkeypatch, file1):
    """
    Test interactive recheck function.
    """
    outpath = str(file1) + ".torrent"
    torrent = torrentclass(path=file1, outfile=outpath)
    filemeta, _ = torrent.write(outfile=outpath)
    seq = ["recheck", filemeta, str(file1)]
    it = iter(seq)
    monkeypatch.setattr(MOCK, lambda *_: next(it))
    result = select_action()
    assert result == 100



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Testing functions for the sub-action commands from command line args.
"""
import io
import os
import sys
import shutil
from argparse import Namespace
from hashlib import sha1, sha256  # nosec
from pathlib import Path
from urllib.parse import quote_plus

import pyben
import pytest

from tests import (
    dir1, dir2, file1, metafile1, metafile2, rmpath, tempfile, torrents)
from torrentfile.cli import execute
from torrentfile.commands import (
    find_config_file, info, magnet, parse_config_file, rebuild, recheck)
from torrentfile.hasher import merkle_root
from torrentfile.utils import ArgumentError


def test_fix():
    """
    Test dir1 fixture is not None.
    """
    assert dir1 and metafile1 and file1 and metafile2 and dir2


def test_magnet_uri(metafile1):
    """
    Test create magnet function digest.
    """
    magnet_link = magnet(metafile1)
    meta = pyben.load(metafile1)
    announce = meta["announce"]
    assert quote_plus(announce) in magnet_link


def test_magnet_hex(metafile1):
    """
    Test create magnet function digest.
    """
    magnet_link = magnet(metafile1)
    meta = pyben.load(metafile1)
    info = meta["info"]
    hashing_func = sha1  # nosec
    if "meta version" in info and "pieces" not in info:
        hashing_func = sha256
    binfo = hashing_func(pyben.dumps(info)).hexdigest()  # nosec
    assert binfo in magnet_link


def test_magnet(metafile1):
    """
    Test create magnet function scheme.
    """
    magnet_link = magnet(metafile1)
    assert magnet_link.startswith("magnet")


def test_magnet_no_announce_list(metafile2):
    """
    Test create magnet function scheme.
    """
    meta = pyben.load(metafile2)
    del meta["announce-list"]
    pyben.dump(meta, metafile2)
    if "meta version" in meta["info"]:
        prefix = "btmh"
    else:
        prefix = "btih"
    magnet_link = magnet(metafile2)
    assert prefix in magnet_link


def test_magnet_no_announce(metafile2):
    """
    Test create magnet function scheme.
    """
    meta = pyben.load(metafile2)
    del meta["announce-list"]
    del meta["announce"]
    pyben.dump(meta, metafile2)
    magnet_link = magnet(metafile2)
    assert magnet_link.startswith("magnet")


def test_magnet_empty():
    """
    Test create magnet function scheme.
    """
    try:
        magnet("file_that_does_not_exist")
    except FileNotFoundError:
        assert True


def test_info_stdout(file1):
    """
    Test the info function properly sends output to stdout.
    """
    outfile = str(file1) + ".torrent"
    args = [
        "torrentfile",
        "create",
        "--announce",
        "url1",
        "--web-seed",
        "url5",
        "url3",
        "--http-seed",
        "url4",
        "--private",
        "-o",
        outfile,
        "--comment",
        "CommentExample",
        "--source",
        "SourceExample",
        str(file1),
    ]
    sys.argv = args
    execute()

    class NameSpace:
        """
        Stand in substitution for argparse.Namespace object.
        """

        metafile = str(file1) + ".torrent"

    stdout = io.StringIO()
    sys.stdout = stdout

    output = info(NameSpace)
    stdout.seek(0)
    written = stdout.read()
    assert output in written


@pytest.mark.parametrize(
    "field",
    ["name", "announce", "source", "comment", "private", "announce-list"],
)
def test_info(field, file1):
    """
    Test the info_command action from the Command Line Interface.
    """
    outfile = str(file1) + ".torrent"
    args = [
        "torrentfile",
        "create",
        "-a",
        "url1",
        "url2",
        "url3",
        "--web-seed",
        "url4",
        "url5",
        "--http-seed",
        "url6",
        "url7",
        "--private",
        "-o",
        outfile,
        "--comment",
        "ExampleComment",
        "--source",
        "examplesource",
        str(file1),
    ]
    sys.argv = args
    execute()

    class Space:
        """
        Stand in substitution for argparse.Namespace object.
        """

        metafile = str(file1) + ".torrent"

    output = info(Space)
    assert field in output


def test_magnet_cli(metafile1):
    """
    Test magnet creation through CLI interface.
    """
    sys.argv[1:] = ["m", str(metafile1)]
    uri = execute()
    assert "magnet" in uri


def test_create_unicode_name(file1):
    """
    Test Unicode information in CLI args.
    """
    parent = os.path.dirname(file1)
    filename = os.path.join(parent, "丂七万丈三与丏丑丒专且丕世丗両丢丣两严丩个丫丬中丮丯.torrent")
    args = [
        "torrentfile",
        "-v",
        "create",
        "-a",
        "tracker_url.com/announce_3456",
        "tracker_url.net/announce_3456",
        "--source",
        "sourcetext",
        "--comment",
        "filename is 丂七万丈三与丏丑丒专且丕世丗両丢丣两严丩个丫丬中丮丯.torrent",
        "-o",
        str(filename),
        str(file1),
    ]
    sys.argv = args
    execute()
    assert os.path.exists(filename)


@pytest.mark.parametrize("blocks", [[], [sha1(b"1010").digest()]])  # nosec
def test_merkle_root_no_blocks(blocks):
    """
    Test running merkle root function with 1 and 0 len lists.
    """
    if blocks:
        assert merkle_root(blocks)
    else:
        assert not merkle_root(blocks)


@pytest.mark.parametrize("torrent", torrents())
def test_mixins_progbar(torrent):
    """
    Test progbar mixins with small file.
    """
    tfile = tempfile(exp=14)
    outfile = str(tfile) + ".torrent"
    msg = "1234abcd" * 80
    with open(tfile, "wb") as temp:
        temp.write(msg.encode("utf-8"))
    args = {
        "path": str(tfile),
        "--prog": "1",
    }
    metafile = torrent(**args)
    output, _ = metafile.write(outfile=outfile)
    assert output == outfile
    rmpath(tfile, outfile)


@pytest.mark.parametrize("torrent", torrents())
def test_mixins_progbar_deep_nesting(torrent):
    """
    Test progbar mixins with small file.
    """
    tfile = tempfile(exp=14)
    dirname = os.path.dirname(tfile)
    nested = "some_super_long_name_to_test_path_length_limits"
    nesting = os.path.join(dirname, nested)
    if not os.path.exists(nesting):
        os.mkdir(nesting)
    tmpfile = os.path.join(nesting, os.path.basename(tfile))
    shutil.move(tfile, tmpfile)
    outfile = str(tmpfile) + ".torrent"
    msg = "1234abcd" * 80
    with open(tmpfile, "wb") as temp:
        temp.write(msg.encode("utf-8"))
    args = {
        "path": str(tmpfile),
        "--prog": "1",
    }
    metafile = torrent(**args)
    output, _ = metafile.write(outfile=outfile)
    assert output == outfile
    rmpath(tmpfile, outfile)


@pytest.fixture
def build(dir2, metafile2):
    """Create fixture for testing rebuild command."""
    basedir = os.path.dirname(dir2)
    parent = os.path.dirname(basedir)
    dest = os.path.join(parent, "dest")
    if os.path.exists(dest):
        rmpath(dest)  # pragma: nocover
    os.mkdir(dest)

    class Namespace:
        """Command line args for rebuild command."""

        metafiles = [os.path.dirname(metafile2)]
        contents = [basedir]
        destination = dest

    yield Namespace
    rmpath(dest)


def test_rebuild(build):
    """Test the rebuild function in the commands module."""
    counter = rebuild(build)
    assert counter > 0


def test_recheck_with_dir():
    """Test running the recheck command with a directory as the metafile."""
    path = os.path.dirname(__file__)

    class Namespace:
        """Emulates the namespace class from argparse module."""

        metafile = path
        content = path

    try:
        recheck(Namespace)
    except ArgumentError:
        assert True


@pytest.fixture(params=[("http-seed", False), ("web-seed", True)])
def config(request):
    """Test config file contents."""
    field, private = request.param
    contents = f"""
[config]
announce =
    url3
    url4
{field} =
    url4
    url5
private = {str(private).lower()}
piece-length = 16
comment = some comment
source = tracker
"""
    return contents


@pytest.fixture(params=[True, False])
def namespace(request, tmp_path):
    """Test fixture for pytest."""
    if request.param:
        ns = Namespace(config=True, config_path=None)
    else:
        ns = Namespace(config=True, config_path=tmp_path / "torrentfile.ini")
    return ns


@pytest.fixture
def configfile(namespace, config):
    """Test fixture for configfile parsing."""
    if namespace.config_path:
        path = namespace.config_path
    else:
        base = Path.home() / ".torrentfile"
        None if os.path.exists(base) else os.mkdir(base)
        path = base / "torrentfile.ini"
    path.write_text(config, encoding="utf8")
    yield namespace
    os.remove(path)


def test_find_config_file(configfile):
    """Test find config file function in commands module."""
    assert find_config_file(configfile) is not None


def test_parse_config_file(configfile):
    """Test parse config file function in commands module."""
    kwargs = {"out": "./somepath"}
    config = find_config_file(configfile)
    parse_config_file(config, kwargs)
    assert "announce" in kwargs


@pytest.mark.parametrize("path", [None, Path.home() / "torrentfile.ini"])
def test_find_config_file_missing(path):
    """Test find config file function with missing config file."""
    ns = Namespace(config=True, config_path=path)
    filename = "torrentfile.ini"
    home = Path.home()
    paths = [
        os.path.join(os.getcwd(), filename),
        home / ".torrentfile" / filename,
        home / ".config" / ".torrentfile" / filename,
    ]
    existing = [i for i in paths if os.path.exists(i)]
    list(map(os.remove, existing))
    try:
        find_config_file(ns)
    except FileNotFoundError:
        assert True



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Clases and functions for the rebuild or reassemble subcommand.

Re-assemble a torrent into the propper directory structure as indicated by a
torrent meta file, and validate the contents of each file allong the
way. Displays a progress bar for each torrent.
"""
import os
import math
import logging
from hashlib import sha1
from pathlib import Path

import pyben

from torrentfile.hasher import HasherV2
from torrentfile.mixins import CbMixin, ProgMixin
from torrentfile.utils import copypath

logger = logging.getLogger(__name__)
SHA1 = 20


class PathNode:
    """
    Base class representing information regarding a file included in torrent.
    """

    def __init__(
        self,
        start: int = None,
        stop: int = None,
        full: str = None,
        filename: str = None,
        path: str = None,
        length: int = None,
    ):
        """
        Hold file information that contributes to the contents of torrent.

        Parameters
        ----------
        start : int, optional
            where the piece starts, by default None
        stop : int, optional
            where the piece ends, by default None
        full : str, optional
            full path, by default None
        filename : str, optional
            filename, by default None
        path : str, optional
            parent path, by default None
        length : int, optional
            size, by default None
        """
        self.path = path
        self.start = start
        self.stop = stop
        self.length = length
        self.filename = filename
        self.full = full

    def get_part(self, path: str) -> bytes:
        """
        Extract the part of the file needed to complete the hash.

        Parameters
        ----------
        path : str
            filesystem path location of file.

        Returns
        -------
        bytes
            part of the file's contents
        """
        with open(path, "rb") as fd:
            if self.start:
                fd.seek(self.start)
            if self.stop != -1:
                partial = fd.read(self.stop - self.start)
            else:
                partial = fd.read()
        return partial

    def __len__(self) -> int:
        """
        Return size of the file.

        Returns
        -------
        int
            total size
        """
        return self.length


class PieceNode:
    """
    Base class representing a single SHA1 hash block of data from a torrent.
    """

    def __init__(self, piece: bytes):
        """
        Store information about an individual SHA1 hash for a torrent file.

        _extended_summary_

        Parameters
        ----------
        piece : bytes
            SHA1 hash bytes
        """
        self.piece = piece
        self.paths = []
        self.result = None
        self.dest = None

    def append(self, pathnode: PathNode):
        """
        Append the path argument to the paths list attribute.

        Parameters
        ----------
        pathnode : PathNode
            the pathnode
        """
        self.paths.append(pathnode)

    def _find_matches(self, filemap: dict, paths: list, data: bytes) -> bool:
        """
        Gather relavent sections of the files in the list and check the hash.

        Parameters
        ----------
        filemap : dict
            dictionary containing filename and path details
        paths : list
            list of pathnodes
        data : bytes
            raw file contents

        Returns
        -------
        bool
            success state
        """
        if not paths:
            piece_hash = sha1(data).digest()  # nosec
            return piece_hash == self.piece
        pathnode = paths[0]
        filename = pathnode.filename
        if filename not in filemap:
            return False  # pragma: nocover
        for loc, size in filemap[filename]:
            if size != len(pathnode):
                continue
            partial = pathnode.get_part(loc)
            val = self._find_matches(filemap, paths[1:], data + partial)
            if val:
                dest_path = os.path.join(self.dest, pathnode.full)
                copypath(loc, dest_path)
            return val
        return False

    def find_matches(self, filemap: dict, dest: str) -> bool:
        """
        Find the matching files for each path in the node.

        Parameters
        ----------
        filemap : dict
            filename and details
        dest : str
            target destination path

        Returns
        -------
        bool
            success status
        """
        self.dest = dest
        self.result = self._find_matches(filemap, self.paths[:], bytes())
        return self.result


class Metadata(CbMixin, ProgMixin):
    """
    Class containing the metadata contents of a torrent file.
    """

    def __init__(self, path: str):
        """
        Construct metadata object for torrent info.

        Parameters
        ----------
        path : str
            path to the .torrent file.
        """
        self.path = os.path.abspath(path)
        self.name = None
        self.piece_length = 1
        self.meta_version = 1
        self.pieces = b""
        self.piece_nodes = []
        self.length = 0
        self.files = []
        self.filenames = set()
        self.extract()
        if self.meta_version == 2:
            self.num_pieces = len(self.filenames)
        else:
            self.num_pieces = math.ceil(len(self.pieces) / SHA1)

    def extract(self):
        """
        Decode and extract information for the .torrent file.
        """
        meta = pyben.load(self.path)
        info = meta["info"]
        self.piece_length = info["piece length"]
        self.name = info["name"]
        self.meta_version = info.get("meta version", 1)
        self.pieces = info.get("pieces", bytes())
        if self.meta_version == 2:
            self._parse_tree(info["file tree"], [self.name])
        elif "length" in info:
            self.length += info["length"]
            self.is_file = True
            self.filenames.add(info["name"])
            self.files.append({
                "path": Path(self.name).parent,
                "filename": self.name,
                "full": self.name,
                "length": self.length,
            })
        elif "files" in info:
            for f in info["files"]:
                path = f["path"]
                full = os.path.join(self.name, *path)
                self.files.append({
                    "path": Path(full).parent,
                    "filename": path[-1],
                    "full": full,
                    "length": f["length"],
                })
                self.length += f["length"]
                self.filenames.add(path[-1])

    def _map_pieces(self):
        """
        Create PathNode and PieceNode details for each piece in the torrent.
        """
        total_pieces = len(self.pieces) // SHA1
        remainder = file_index = 0
        current = {}
        for i in range(total_pieces):
            begin = SHA1 * i
            piece = PieceNode(self.pieces[begin:begin + SHA1])
            target = self.piece_length
            if remainder:
                start = current["length"] - remainder
                if remainder < target:
                    stop = -1
                    target -= remainder
                    remainder = 0
                    file_index += 1
                else:
                    stop = start + target
                    remainder -= target
                    target -= target
                pathnode = PathNode(start=start, stop=stop, **current)
                piece.append(pathnode)
            while target > 0 and file_index < len(self.files):
                start = 0
                current = self.files[file_index]
                size = current["length"]
                if size < target:
                    stop = -1
                    target -= size
                    file_index += 1
                else:
                    stop = target
                    remainder = size - target
                    target = 0
                pathnode = PathNode(start=start, stop=stop, **current)
                piece.append(pathnode)
            self.piece_nodes.append(piece)

    def _parse_tree(self, tree: dict, partials: list):
        """
        Parse the file tree dictionary of the torrent metafile.

        Parameters
        ----------
        tree : dict
            the dictionary representation of a file tree.
        partials : list
            list of paths leading up to the current key value.
        """
        for key, val in tree.items():
            if "" in val:
                self.filenames.add(key)
                path = Path(os.path.join(*partials))
                full = Path(os.path.join(path, key))
                length = val[""]["length"]
                root = val[""]["pieces root"]
                self.files.append({
                    "path": path,
                    "full": full,
                    "filename": key,
                    "length": length,
                    "root": root,
                })
                self.length += length
            else:
                self._parse_tree(val, partials + [key])

    def _match_v1(self, filemap: dict, dest: str):
        """
        Check each of the nodes against the filemap dictionary for matches.

        Parameters
        ----------
        filemap : dict
            filenames and filesystem information
        dest : str
            target destination path
        """
        self._map_pieces()
        copied = []
        for piece_node in self.piece_nodes:
            paths = piece_node.paths
            if len(paths) == 1 and paths[0].path in copied:
                self._update()
                continue
            if piece_node.find_matches(filemap, dest):
                for pathnode in paths:
                    if pathnode.path not in copied:
                        copied.append(pathnode.path)
                        dest_path = os.path.join(dest, pathnode.path)
                        self._update()
                        self.cb(pathnode.path, dest_path, self.num_pieces)

    def _match_v2(self, filemap: dict, dest: str):
        """
        Rebuild method for torrent v2 files.

        Parameters
        ----------
        filemap : dict
            filesystem information
        dest : str
            destiantion path
        """
        for entry in self.files:
            filename = entry["filename"]
            length = entry["length"]
            if filename not in filemap:
                continue  # pragma: nocover
            paths = filemap[filename]
            for path, size in paths:
                if size == length:
                    hasher = HasherV2(path, self.piece_length, True)
                    if entry["root"] == hasher.root:
                        dest_path = os.path.join(dest, entry["full"])
                        copypath(entry["path"], dest_path)
                        self._update()
                        self.cb(path, dest_path, self.num_pieces)
                        break

    def rebuild(self, filemap: dict, dest: str):
        """
        Rebuild torrent file contents from filemap at dest.

        Searches through the contents of the meta file and compares filenames
        with those in the filemap dict, and if found checks their contents,
        and copies them to the destination path.

        Parameters
        ----------
        filemap : dict
            filesystem information
        dest : str
            destiantion path
        """
        self._prog = None
        if self.meta_version == 2:
            self._match_v2(filemap, dest)
        else:
            self._match_v1(filemap, dest)
        if self._prog is not None:
            self.prog_close()

    def _update(self):
        """Start and updating the progress bar."""
        if self._prog is None:
            self._prog = True
            self.prog_start(self.num_pieces, self.name, unit="piece")
        self.prog_update(1)


class Assembler(CbMixin):
    """
    Does most of the work in attempting the structure of torrentfiles.

    Requires three paths as arguments.
    - torrent metafile or directory containing multiple meta files
    - directory containing the contents of meta file
    - directory where torrents will be re-assembled
    """

    def __init__(self, metafiles: list, contents: list, dest: str):
        """
        Reassemble given torrent file from given cli arguments.

        Rebuild metafiles and contents into their original directory
        structure as much as possible in the destination directory.
        Takes two paths as parameters,
        - file or directory containing 1 or more torrent meta files
        - path to where the contents are belived to be located.

        Parameters
        ----------
        metafiles : str
            path to torrent metafile or directory containing torrent metafiles.
        contents : str
            path to content or directory containing content that belongs to
            torrentfile.
        dest: str
            path to the directory where rebuild will take place.
        """
        Metadata.set_callback(self._callback)
        self.counter = 0
        self._lastlog = None
        self.contents = contents
        self.dest = dest
        self.meta_paths = metafiles
        self.metafiles = self._get_metafiles()
        filenames = set()
        for meta in self.metafiles:
            filenames |= meta.filenames
        self.filemap = _index_contents(self.contents, filenames)

    def _callback(self, filename: str, dest: str, num_pieces: int):
        """
        Run the callback functions associated with Mixin for copied files.

        Parameters
        ----------
        filename : str
            filename
        dest : str
            destination path
        num_pieces : int
            number of hash pieces
        """
        self.counter += 1
        message = f"Matched:{num_pieces} {filename} -> {dest}"
        if message != self._lastlog:
            self._lastlog = message
            logger.debug(message)

    def assemble_torrents(self):
        """
        Assemble collection of torrent files into original structure.

        Returns
        -------
        int
            number of files copied
        """
        for metafile in self.metafiles:
            logger.info("#%s Searching contents for %s", self.counter,
                        metafile.name)
            self.rebuild(metafile)
        return self.counter

    def rebuild(self, metafile: Metadata) -> None:
        """
        Build the torrent file structure from contents of directory.

        Traverse contents dir and compare discovered files
        with files listed in torrent metadata and copy
        the matches to the destination directory respecting folder
        structures along the way.
        """
        metafile.rebuild(self.filemap, self.dest)

    def _iter_files(self, path: str) -> list:
        """
        Iterate through metfiles directory createing Metafile objects.

        Parameters
        ----------
        path : str
            fs path

        Returns
        -------
        list
            list of Metadata Object
        """
        metafiles = []
        for filename in os.listdir(path):
            if filename.lower().endswith(".torrent"):
                try:
                    meta = Metadata(os.path.join(path, filename))
                    metafiles.append(meta)
                except ValueError:  # pragma: nocover
                    self.counter -= 1
        return metafiles

    def _get_metafiles(self) -> list:
        """
        Collect all .torrent meta files from given directory or file.

        Returns
        -------
        list
            metafile objects
        """
        metafiles = []
        for path in self.meta_paths:
            if os.path.exists(path):
                if os.path.isdir(path):
                    metafiles += self._iter_files(path)
                elif path.lower().endswith(".torrent"):
                    meta = Metadata(path)
                    metafiles.append(meta)
        return metafiles


def _index_contents(contents: list, filenames: set) -> dict:
    """
    Collect all of the filenames and their respective paths.

    Parameters
    ----------
    contents : list
        paths to traverse looking for filenames
    filenames : set
        set of filenames to look for

    Returns
    -------
    dict
        all filenames and their respective paths.
    """
    mapping = {}
    for dirpath in contents:
        mapped = _index_content(dirpath, filenames)
        for key, value in mapped.items():
            mapping.setdefault(key, [])
            mapping[key].extend(value)
    return mapping


def _index_content(root: str, filenames: set) -> dict:
    """
    Collect filenames from directory or file.

    Parameters
    ----------
    root : str
        path to search for filenames
    filenames : set
        set of filenames to search for

    Returns
    -------
    dict
        filenames and their respective paths
    """
    filemap = {}
    if os.path.isfile(root):
        name = os.path.basename(root)
        if name in filenames:
            size = os.path.getsize(root)
            filemap.setdefault(name, [])
            filemap[name].append((root, size))
        return filemap
    if os.path.isdir(root):
        for path in os.listdir(root):
            fullpath = os.path.join(root, path)
            resultmap = _index_content(fullpath, filenames)
            for key, value in resultmap.items():
                filemap.setdefault(key, [])
                filemap[key].extend(value)
    return filemap



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Utility functions and classes used throughout package.
"""

import os
import math
import shutil
from pathlib import Path


class Memo:
    """
    Memoize cache.

    Parameters
    ----------
    func : Callable
        The results of this callable will be cached.
    """

    def __init__(self, func):
        """
        Construcor for cache.
        """
        self.func = func
        self.counter = 0
        self.cache = {}

    def __call__(self, path: str):
        """
        Invoke each time memo function is called.

        Parameters
        ----------
        path : str
            The relative or absolute path being used as key in cache dict.

        Returns
        -------
        Any :
            The results of calling the function with path.
        """
        if path in self.cache and os.path.exists(path):
            self.counter += 1
            return self.cache[path]
        result = self.func(path)
        self.cache[path] = result
        return result


class MissingPathError(Exception):
    """
    Path parameter is required to specify target content.

    Creating a .torrent file with no contents seems rather silly.

    Parameters
    ----------
    message : str
        Message for user (optional).
    """

    def __init__(self, message: str = None):
        """
        Raise when creating a meta file without specifying target content.

        The `message` argument is a message to pass to Exception base class.
        """
        self.message = f"Path arguement is missing and required {str(message)}"
        super().__init__(message)


class PieceLengthValueError(Exception):
    """
    Piece Length parameter must equal a perfect power of 2.

    Parameters
    ----------
    message : str
        Message for user (optional).
    """

    def __init__(self, message: str = None):
        """
        Raise when creating a meta file with incorrect piece length value.

        The `message` argument is a message to pass to Exception base class.
        """
        self.message = f"Incorrect value for piece length: {str(message)}"
        super().__init__(message)


class ArgumentError(Exception):
    """
    Exception for mismatched or mistyped CLI arguments.
    """


SUFFIXES = ["KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"]


def humanize_bytes(amount: int) -> str:
    """
    Convert integer into human readable memory sized denomination.

    Parameters
    ----------
    amount : int
        total number of bytes.

    Returns
    -------
    str
        human readable representation of the given amount of bytes.
    """
    base = 1024
    amount = float(amount)
    value = abs(amount)
    if value == 1:
        return f"{amount} Byte"
    if value < base:
        return f"{amount} Bytes"
    for i, s in enumerate(SUFFIXES):
        unit = base**(i + 2)
        if value < unit:
            break
    value = base * amount / unit
    return f"{value:.1f} {s}"


def normalize_piece_length(piece_length: int) -> int:
    """
    Verify input piece_length is valid and convert accordingly.

    Parameters
    ----------
    piece_length : int | str
        The piece length provided by user.

    Returns
    -------
    int
        normalized piece length.

    Raises
    ------
    PieceLengthValueError :
        Piece length is improper value.
    """
    if isinstance(piece_length, str):
        if piece_length.isnumeric():
            piece_length = int(piece_length)
        else:
            raise PieceLengthValueError(piece_length)

    if piece_length > (1 << 14):
        if 2**math.log2(piece_length) == piece_length:
            return piece_length
        raise PieceLengthValueError(piece_length)

    if 13 < piece_length < 26:
        return 2**piece_length
    if piece_length <= 13:
        raise PieceLengthValueError(piece_length)

    log = int(math.log2(piece_length))
    if 2**log == piece_length:
        return piece_length
    raise PieceLengthValueError


def get_piece_length(size: int) -> int:
    """
    Calculate the ideal piece length for bittorrent data.

    Parameters
    ----------
    size : int
        Total bits of all files incluided in .torrent file.

    Returns
    -------
    int
        Ideal piece length.
    """
    exp = 14
    while size / (2**exp) > 1000 and exp < 24:
        exp += 1
    return 2**exp


@Memo
def filelist_total(pathstring: str) -> os.PathLike:
    """
    Perform error checking and format conversion to os.PathLike.

    Parameters
    ----------
    pathstring : str
        An existing filesystem path.

    Returns
    -------
    os.PathLike
        Input path converted to bytes format.

    Raises
    ------
    MissingPathError
        File could not be found.
    """
    if os.path.exists(pathstring):
        path = Path(pathstring)
        return _filelist_total(path)
    raise MissingPathError


def _filelist_total(path: os.PathLike) -> tuple:
    """
    Recursively search directory tree for files.

    Parameters
    ----------
    path : str
        Path to file or directory base

    Returns
    -------
    Tuple[int, List] :
        int - sum of sizes for all files collected
        list - all file paths within directory tree
    """
    if path.is_file():
        file_size = os.path.getsize(path)
        return file_size, [str(path)]
    total = 0
    filelist = []
    if path.is_dir():
        for item in path.iterdir():
            size, paths = filelist_total(item)
            total += size
            filelist.extend(paths)
    return total, sorted(filelist)


def path_size(path: str) -> int:
    """
    Return the total size of all files in path recursively.

    Parameters
    ----------
    path : str
        path to target file or directory.

    Returns
    -------
    int
        total size of files.
    """
    total_size, _ = filelist_total(path)
    return total_size


def get_file_list(path: str) -> list:
    """
    Return a sorted list of file paths contained in directory.

    Parameters
    ----------
    path : str
        target file or directory.

    Returns
    -------
    list :
        sorted list of file paths.
    """
    _, filelist = filelist_total(path)
    return filelist


def path_stat(path: str) -> tuple:
    """
    Calculate directory statistics.

    Parameters
    ----------
    path : str
        The path to start calculating from.

    Returns
    -------
    Tuple[list, int, int] :
        list - List of all files contained in Directory
        int - Total sum of bytes from all contents of dir
        int - The size of pieces of the torrent contents.
    """
    total_size, filelist = filelist_total(path)
    piece_length = get_piece_length(total_size)
    return (filelist, total_size, piece_length)


def path_piece_length(path: str) -> int:
    """
    Calculate piece length for input path and contents.

    Parameters
    ----------
    path : str
        The absolute path to directory and contents.

    Returns
    -------
    int
        The size of pieces of torrent content.
    """
    psize = path_size(path)
    return get_piece_length(psize)


def next_power_2(value: int) -> int:
    """
    Calculate the next perfect power of 2 equal to or greater than value.

    Parameters
    ----------
    value : int
        integer value that is less than some perfect power of 2.

    Returns
    -------
    int
        The next power of 2 greater than value, or value if already power of 2.
    """
    if not value & (value - 1) and value:
        return value
    start = 1
    while start < value:
        start <<= 1
    return start


def copypath(source: str, dest: str) -> None:
    """
    Copy the file located at source to dest.

    If one or more directory paths don't exist in dest, they will be created.
    If dest already exists and dest and source are the same size, it will be
    ignored, however if dest is smaller than source, dest will be overwritten.

    Parameters
    ----------
    source : str
        path to source file
    dest : str
        path to target destination
    """
    if not os.path.exists(source) or (os.path.exists(dest)
                                      and os.path.getsize(source)
                                      <= os.path.getsize(dest)):
        return
    path_parts = Path(dest).parts
    if len(path_parts) > 1:
        root = path_parts[0]
        path_parts = path_parts[1:-1]
        if not os.path.exists(root):
            os.mkdir(root)  # pragma: nocover
        for part in path_parts:
            path = os.path.join(root, part)
            if not os.path.exists(path):
                os.mkdir(path)
            root = path
        shutil.copy(source, dest)


def toggle_debug_mode(switch_on: bool):
    """
    Switch the environment variable debug indicator on or off.

    Parameters
    ----------
    switch_on : bool
        if true turn debug mode on otherwise off
    """
    os.environ["TORRENTFILE_DEBUG"] = "ON" if switch_on else "OFF"


def debug_is_on() -> bool:
    """
    Return True if debug mode is on in environment variables.

    Returns
    -------
    bool
        is debug mode on
    """
    return os.environ["TORRENTFILE_DEBUG"] == "ON"


def check_path_writable(path: str) -> bool:
    """
    Test if output path is writable.

    Parameters
    ----------
    path : str
        file system path string

    Returns
    -------
    bool
        True if writeable, otherwise raises PermissionError
    """
    try:
        if path.endswith("\\") or path.endswith("/"):
            path = os.path.join(path, ".torrent")
        with open(path, "ab") as _:
            pass
        os.remove(path)
    except PermissionError as err:  # pragma: nocover
        directory = os.path.dirname(path)
        message = f"Target directory is not writeable {directory}"
        raise PermissionError(message) from err
    return True



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
The commands module contains the Action Commands executed by the CLI script.

Each function pertains to a command line action/subcommand and drives specific
features of the application.

Functions
---------
- create
- info
- edit
- recheck
- magnet
- rebuild
- find_config_file
- parse_config_file
- get_magnet
"""

import os
import sys
import shutil
import logging
import configparser
from argparse import Namespace
from hashlib import sha1, sha256
from pathlib import Path
from urllib.parse import quote_plus

import pyben

from torrentfile.edit import edit_torrent
from torrentfile.interactive import select_action
from torrentfile.rebuild import Assembler
from torrentfile.recheck import Checker
from torrentfile.torrent import TorrentAssembler, TorrentFile
from torrentfile.utils import ArgumentError, check_path_writable

logger = logging.getLogger(__name__)


def find_config_file(args: Namespace) -> str:
    """
    Locate the path to the torrentfile configuration file.

    Parameters
    ----------
    args : Namespace
        command line argument values

    Returns
    -------
    str
        path to the configuration file

    Raises
    ------
    FileNotFoundError
        raised if configuration file not found.
    """
    path = None
    error_message = "Could not find configuration file."
    if args.config_path:
        if os.path.exists(args.config_path):
            path = args.config_path
        else:
            raise FileNotFoundError(error_message)
    else:
        filename = "torrentfile.ini"
        paths = [
            os.path.join(os.getcwd(), filename),
            Path.home() / ".torrentfile" / filename,
            Path.home() / ".config" / ".torrentfile" / filename,
        ]
        for subpath in paths:
            if os.path.exists(subpath):
                path = subpath
                break
    if path is None:
        raise FileNotFoundError(error_message)
    return path


def parse_config_file(path: str, kwargs: dict):
    """
    Parse configuration file for torrent setup details.

    Parameters
    ----------
    path : str
        path to configuration file
    kwargs : dict
        options from command line arguments
    """
    config = configparser.ConfigParser()
    config.read(path)

    for key, val in config["config"].items():
        if key.lower() in ["announce", "http-seed", "web-seed", "tracker"]:
            val = [i for i in val.split("\n") if i]

            if key.lower() == "http-seed":
                kwargs["httpseeds"] = val

            elif key.lower() == "web-seed":
                kwargs.setdefault("url-list", [])
                kwargs["url-list"] = val

            else:
                kwargs[key.lower()] = val

        elif key.lower() == "piece-length":
            kwargs["piece_length"] = val

        elif key.lower() == "meta-version":
            kwargs["meta_version"] = val

        elif val.lower() == "true":
            kwargs[key.lower()] = True

        elif val.lower() == "false":
            kwargs[key.lower()] = False

        else:
            kwargs[key.lower()] = val


def create(args: Namespace) -> Namespace:
    """
    Execute the create CLI sub-command to create a new torrent metafile.

    Parameters
    ----------
    args : Namespace
        positional and optional CLI arguments.

    Returns
    -------
    torrentfile.MetaFile
        object containing the path to created metafile and its contents.
    """
    kwargs = vars(args)
    if args.config:
        path = find_config_file(args)
        parse_config_file(path, kwargs)  # pragma: nocover

    if args.outfile:
        check_path_writable(args.outfile)

    else:  # pragma: nocover
        samplepath = os.path.join(os.getcwd(), ".torrent")
        check_path_writable(samplepath)

    logger.debug("Creating torrent from %s", args.content)
    if args.meta_version == "1":
        torrent = TorrentFile(**kwargs)

    else:
        torrent = TorrentAssembler(**kwargs)
    outfile, meta = torrent.write()

    if args.magnet:
        magnet(outfile, version=0)

    args.torrent = torrent
    args.kwargs = kwargs
    args.outfile = outfile
    args.meta = meta

    print("\nTorrent Save Path: ", os.path.abspath(str(outfile)))
    logger.debug("Output path: %s", str(outfile))
    return args


def info(args: Namespace) -> str:
    """
    Show torrent metafile details to user via stdout.

    Prints full details of torrent file contents to the terminal in
    a clean and readable format.

    Parameters
    ----------
    args : dict
        command line arguements provided by the user.

    Returns
    -------
    str
        The output printed to the terminal.
    """
    metafile = args.metafile
    meta = pyben.load(metafile)
    data = meta["info"]
    del meta["info"]

    meta.update(data)
    if "private" in meta and meta["private"] == 1:
        meta["private"] = "True"

    if "announce-list" in meta:
        lst = meta["announce-list"]
        meta["announce-list"] = ", ".join([j for i in lst for j in i])

    if "url-list" in meta:
        meta["url-list"] = ", ".join(meta["url-list"])

    if "httpseeds" in meta:
        meta["httpseeds"] = ", ".join(meta["httpseeds"])

    text = []
    longest = max(len(i) for i in meta.keys())

    for key, val in meta.items():
        if key not in ["pieces", "piece layers", "files", "file tree"]:
            prefix = longest - len(key) + 1
            string = key + (" " * prefix) + str(val)
            text.append(string)

    most = max(len(i) for i in text)
    text = ["-" * most, "\n"] + text + ["\n", "-" * most]
    output = "\n".join(text)
    sys.stdout.write(output)
    sys.stdout.flush()
    return output


def edit(args: Namespace) -> str:
    """
    Execute the edit CLI sub-command with provided arguments.

    Provides functionality that can change the details of a torrentfile
    that preserves all of the hash piece information so as not to break
    the torrentfile.

    Parameters
    ----------
    args : Namespace
        positional and optional CLI arguments.

    Returns
    -------
    str
        path to edited torrent file.
    """
    metafile = args.metafile
    logger.info("Editing %s Meta File", str(args.metafile))

    editargs = {
        "url-list": args.url_list,
        "httpseeds": args.httpseeds,
        "announce": args.announce,
        "source": args.source,
        "private": args.private,
        "comment": args.comment,
    }
    return edit_torrent(metafile, editargs)


def recheck(args: Namespace) -> str:
    """
    Execute recheck CLI sub-command.

    Checks the piece hashes within a pre-existing torrent file
    and does a piece by piece check with the contents of a file
    or directory for completeness and validation.

    Parameters
    ----------
    args : Namespace
        positional and optional arguments.

    Returns
    -------
    str
        The percentage of content currently saved to disk.
    """
    metafile = args.metafile
    content = args.content

    if os.path.isdir(metafile):
        raise ArgumentError(f"Error: Unable to parse directory {metafile}. "
                            "Check the order of the parameters.")

    logger.debug("Validating %s <---------------> %s contents", metafile,
                 content)

    msg = f"Rechecking  {metafile} ...\n"
    halfterm = shutil.get_terminal_size().columns / 2
    padding = int(halfterm - (len(msg) / 2)) * " "
    sys.stdout.write(padding + msg)

    checker = Checker(metafile, content)
    logger.debug("Completed initialization of the Checker class")
    result = checker.results()

    message = f"{content} <- {result}% -> {metafile}"
    padding = int(halfterm - (len(message) / 2)) * " "
    sys.stdout.write(padding + message + "\n")
    sys.stdout.flush()
    return result


def rename(args: Namespace) -> str:
    """
    Rename a torrent file to it's original name found in metadata.

    Parameters
    ----------
    args: Namespace
        cli arguments

    Returns
    -------
    str
        renamed file path
    """
    target = args.target
    if not target or not os.path.exists(target):
        raise FileNotFoundError  # pragma: nocover
    meta = pyben.load(target)
    name = meta["info"]["name"]
    parent = os.path.dirname(target)
    new_path = os.path.join(parent, name + ".torrent")
    if os.path.exists(new_path):
        raise FileExistsError  # pragma: nocover
    os.rename(target, new_path)
    return new_path


def get_magnet(namespace: Namespace) -> str:
    """
    Prepare option parameters for retreiving magnet URI.

    Parameters
    ----------
    namespace: Namespace
        command line argument options

    Returns
    -------
    str
        Magnet URI
    """
    metafile = namespace.metafile
    version = int(namespace.meta_version)
    return magnet(metafile, version=version)


def magnet(metafile: str, version: int = 0) -> str:
    """
    Create a magnet URI from a Bittorrent meta file.

    Parameters
    ----------
    metafile : str
        path to bittorrent file
    version: int
        version of bittorrent protocol [default=1]

    Returns
    -------
    str
        Magnet URI
    """
    if not os.path.exists(metafile):
        raise FileNotFoundError(f"No Such File {metafile}")
    meta = pyben.load(metafile)
    info_dict = meta["info"]

    magnet = "magnet:?"
    bencoded_info = pyben.dumps(info_dict)

    v1 = False
    if "meta version" not in info_dict or (version in [1, 3, 0]
                                           and "pieces" in info_dict):
        infohash = sha1(bencoded_info).hexdigest()  # nosec
        magnet += "xt=urn:btih:" + infohash
        v1 = True

    if "meta version" in info_dict and version != 1:
        infohash = sha256(bencoded_info).hexdigest()
        if v1:
            magnet += "&"
        magnet += "xt=urn:btmh:1220" + infohash

    magnet += "&dn=" + quote_plus(info_dict["name"])

    if "announce-list" in meta:
        announce_args = [
            "&tr=" + quote_plus(url) for urllist in meta["announce-list"]
            for url in urllist
        ]
    elif "announce" in meta:
        announce_args = ["&tr=" + quote_plus(meta["announce"])]
    else:
        announce_args = [""]

    trackers = "".join(announce_args)

    magnet += trackers if trackers != "&tr=" else ""

    logger.info("Created Magnet URI %s", magnet)
    sys.stdout.write("\n" + magnet + "\n")
    return magnet


def rebuild(args: Namespace) -> int:
    """
    Attempt to rebuild a torrent based on the a torrent file.

    Recursively look through a directory for files that belong in
    a given torrent file, and rebuild as much of the torrent file
    as possible. Currently only checks if the filename and file
    size are a match.

    1. Check file hashes to improve accuracy

    Parameters
    ----------
    args : Namespace
        command line arguments including the paths neccessary

    Returns
    -------
    int
        total number of content files copied to the rebuild directory
    """
    metafiles = args.metafiles
    dest = args.destination
    contents = args.contents
    for path in [*metafiles, *contents]:
        if not os.path.exists(path):
            raise FileNotFoundError(path)
    assembler = Assembler(metafiles, contents, dest)
    return assembler.assemble_torrents()


interactive = select_action  # for clean import system



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Unittest functions for testing torrentfile utils module.
"""
import math

import pytest

from tests import dir1, dir2, rmpath
from torrentfile import utils
from torrentfile.__main__ import main


def test_main_exists():
    """
    Test if main exists
    """
    assert main


@pytest.mark.parametrize("size", [156634528, 2**30, 67987, 16384, 8563945])
def test_get_piece_length(size):
    """
    Test function for best piece length for given size.
    """
    value = utils.get_piece_length(size)
    assert value % 1024 == 0


@pytest.mark.parametrize("size", [156634528, 2**30, 67987, 16384, 8563945])
def test_get_piece_length_max(size):
    """
    Test function for best piece length for given size maximum.
    """
    value = utils.get_piece_length(size)
    assert value < 2**27


@pytest.mark.parametrize("size", [156634528, 2**30, 67987, 16384, 8563945])
def test_get_piece_length_min(size):
    """
    Test function for best piece length for given size minimum.
    """
    value = utils.get_piece_length(size)
    assert value >= 2**14


def test_get_path_length_mod(dir1):
    """
    Test function for the best piece length for provided path.
    """
    assert utils.path_piece_length(dir1) % (2**14) == 0


def test_get_path_length_min(dir1):
    """
    Test function for getting piece length for folders min.
    """
    assert utils.path_piece_length(dir1) >= (2**14)


def test_get_path_length_max(dir1):
    """
    Test function for getting piece length for folders max.
    """
    assert utils.path_piece_length(dir1) <= (2**27)


def test_path_stat(dir1):
    """
    Test function for acquiring piece length information on folder.
    """
    _, _, piece_length = utils.path_stat(dir1)
    assert piece_length % (2**14) == 0


def test_path_stat_size(dir1):
    """
    Test function for acquiring total size information on folder.
    """
    _, totalsize, _ = utils.path_stat(dir1)
    assert totalsize == (2**18) * 8


def test_path_stat_filelist_size(dir1):
    """
    Test function for acquiring file list information on folder.
    """
    filelist, _, _ = utils.path_stat(dir1)
    assert len(filelist) == 8


def test_get_filelist(dir1):
    """
    Test function for get a list of files in a directory.
    """
    filelist = utils.get_file_list(dir1)
    assert len(filelist) == 8


def test_get_path_size(dir1):
    """
    Test function for getting total size of directory.
    """
    pathsize = utils.path_size(dir1)
    assert pathsize == (2**18) * 8


def test_filelist_total(dir1):
    """
    Test function for acquiring a filelist for directory.
    """
    total, _ = utils.filelist_total(dir1)
    assert total == (2**18) * 8


def test_piecelength_error_fixtures():
    """
    Test exception for uninterpretable piece length value.
    """
    try:
        raise utils.PieceLengthValueError("message")
    except utils.PieceLengthValueError:
        assert True
        assert dir1


def test_missing_path_error():
    """
    Test exception for missing path parameter.
    """
    try:
        raise utils.MissingPathError("message")
    except utils.MissingPathError:
        assert True
        assert dir2


@pytest.mark.parametrize("value", [5, 32, 18, 225, 16384, 256000])
def test_next_power_2(value):
    """
    Test next power of 2 function in utils module.
    """
    result = utils.next_power_2(value)
    log = math.log2(result)
    assert log == int(log)
    assert result % 2 == 0
    assert result >= value


@pytest.mark.parametrize(
    "amount, result",
    [
        (1, f"{float(1)} Byte"),
        (100, "100.0 Bytes"),
        (1100, "1.1 KiB"),
        (1_100_000, "1.0 MiB"),
        (1_100_000_000, "1.0 GiB"),
        (4_400_120_000, "4.1 GiB"),
        (4_000_120_000, "3.7 GiB"),
    ],
)
def test_humanize_bytes(amount, result):
    """
    Test humanize bytes function.
    """
    assert utils.humanize_bytes(amount) == result


@pytest.mark.parametrize("amount, result", [(i, 2**i) for i in range(14, 25)])
def test_normalize_piece_length_int(amount, result):
    """Test normalize piece length function.

    Parameters
    ----------
    amount : `str` or `int`
        piece length or representation
    result : int
        expected output.
    """
    assert utils.normalize_piece_length(amount) == result


@pytest.mark.parametrize("amount, result",
                         [(str(i), 2**i) for i in range(14, 21)])
def test_normalize_piece_length_str(amount, result):
    """Test normalize piece length function.

    Parameters
    ----------
    amount : `str` or `int`
        piece length or representation
    result : int
        expected output.
    """
    assert utils.normalize_piece_length(amount) == result


@pytest.mark.parametrize("amount",
                         ["hello", 11, 0, 100000, 28, "zero", "fifteen"])
def test_norm_plength_errors(amount):
    """Test function to normalize piece length errors.

    Parameters
    ----------
    amount : any
        arguments intended to raise an exception.
    """
    try:
        assert utils.normalize_piece_length(amount)
    except utils.PieceLengthValueError:
        assert True


def test_filelisttotal_missing(dir2):
    """Test function filelist total with missing path.

    Parameters
    ----------
    dir2 : pytest.fixture
        fixture containing a temporary directory
    """
    rmpath(dir2)
    try:
        utils.filelist_total(dir2)
    except utils.MissingPathError:
        assert True


def test_argument_error():
    """
    Test Argument Error.

    Raises
    ------
    utils.ArgumentError
        Arg error
    """
    try:
        raise utils.ArgumentError("This message raised by argument error")
    except utils.ArgumentError:
        assert True



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Torrentfile can create Bittorrent metafiles for any content.

All Bittorrent versions are fully supported.  Torrentfile can also edit,
recheck torrents contents, and create magnet URI's for torrent files.
"""
from torrentfile.cli import execute, main
from torrentfile.commands import create, edit, info, magnet, recheck
from torrentfile.utils import toggle_debug_mode
from torrentfile.version import __version__

toggle_debug_mode(False)

VERSION = __version__

__all__ = [
    "execute",
    "create",
    "edit",
    "info",
    "magnet",
    "recheck",
    "main",
    "VERSION",
]



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Testing functions for the torrent module.
"""
import os

import pytest

from tests import dir1, dir2, rmpath, tempfile, torrents
from torrentfile.mixins import ProgMixin, waiting
from torrentfile.torrent import MetaFile
from torrentfile.utils import MissingPathError


def test_fixtures():
    """
    Test pytest fixtures.
    """
    assert dir1 and dir2


@pytest.mark.parametrize("version", torrents())
def test_torrentfile_missing_path(version):
    """
    Test missing path error exception.
    """
    try:
        version()
    except MissingPathError:
        assert True


def test_metafile_assemble(dir1):
    """
    Test assembling base metafile exception.
    """
    metafile = MetaFile(path=dir1)
    try:
        metafile.assemble()
    except NotImplementedError:
        assert True


@pytest.mark.parametrize("version", torrents())
def test_torrentfile_one_empty(dir2, version):
    """
    Test creating a torrent meta file with given directory plus extra.
    """
    a = next(os.walk(dir2))
    if len(a[-1]) > 0:
        with open(os.path.join(a[0], a[-1][0]), "w", encoding="utf-8") as _:
            pass
    args = {
        "path": dir2,
        "comment": "somecomment",
        "announce": "announce",
    }
    torrent = version(**args)
    assert torrent.meta["announce"] == "announce"


@pytest.mark.parametrize("version", torrents())
def test_torrentfile_extra(dir2, version):
    """
    Test creating a torrent meta file with given directory plus extra.
    """

    def walk(item):
        """
        Edit files in directory structure.
        """
        if item.is_file():
            with open(item, "ab") as binfile:
                binfile.write(bytes(1000))
        elif item.is_dir():
            for sub in item.iterdir():
                walk(sub)

    walk(dir2)
    args = {
        "path": dir2,
        "comment": "somecomment",
        "announce": "announce",
    }
    torrent = version(**args)
    assert torrent.meta["announce"] == "announce"


@pytest.mark.parametrize("num", list(range(17, 25)))
@pytest.mark.parametrize("piece_length", [2**i for i in range(14, 18)])
@pytest.mark.parametrize("version", torrents())
def test_torrentfile_single(version, num, piece_length, capsys):
    """
    Test creating a torrent file from a single file contents.
    """
    tfile = tempfile(exp=num)
    with capsys.disabled():
        version.set_callback(print)
    outfile = str(tfile) + ".torrent"
    args = {
        "path": tfile,
        "comment": "somecomment",
        "announce": "announce",
        "piece_length": piece_length,
        "outfile": outfile,
    }
    trent = version(**args)
    trent.write()
    assert os.path.exists(outfile)
    rmpath(tfile, str(tfile) + ".torrent")


@pytest.mark.parametrize("size", list(range(17, 25)))
@pytest.mark.parametrize("piece_length", [2**i for i in range(14, 18)])
@pytest.mark.parametrize("version", torrents())
def test_torrentfile_single_extra(version, size, piece_length):
    """
    Test creating a torrent file from a single file contents plus extra.
    """
    tfile = tempfile(exp=size)
    with open(tfile, "ab") as binfile:
        binfile.write(bytes(str(tfile).encode("utf-8")))
    outfile = str(tfile) + ".torrent"
    args = {
        "path": tfile,
        "comment": "somecomment",
        "announce": "announce",
        "piece_length": piece_length,
        "outfile": outfile,
    }
    torrent = version(**args)
    torrent.write()
    assert os.path.exists(outfile)
    rmpath(tfile, outfile)


@pytest.mark.parametrize("sze", list(range(17, 25)))
@pytest.mark.parametrize("piecelength", [2**i for i in range(14, 18)])
@pytest.mark.parametrize("ver", torrents())
def test_torrentfile_single_under(ver, sze, piecelength):
    """
    Test creating a torrent file from less than a single file contents.
    """
    tfile = tempfile(exp=sze)
    with open(tfile, "rb") as binfile:
        data = binfile.read()
    with open(tfile, "wb") as binfile:
        binfile.write(data[:-(2**9)])
    outfile = str(tfile) + ".torrent"
    kwargs = {
        "path": tfile,
        "comment": "somecomment",
        "announce": "announce",
        "piece_length": piecelength,
        "outfile": outfile,
    }
    torrent = ver(**kwargs)
    outfile, _ = torrent.write()
    assert os.path.exists(outfile)
    rmpath(tfile, outfile)


def test_create_cwd_fail():
    """Test cwd argument with create command failure."""

    class SuFile:
        """A mock admin file."""

        @staticmethod
        def __fspath__():
            raise PermissionError

        def __str__(self):
            return "SuFile"

    tfile = tempfile()
    torrent = MetaFile(path=tfile)
    sufile = SuFile()
    try:
        assert torrent.write(outfile=sufile)
    except PermissionError:
        assert True
    rmpath(tfile)


def test_waiting_mixin():
    """
    Test waiting function.
    """
    msg = "Testing message"
    lst = []
    timeout = 3
    waiting(msg, lst, timeout=timeout)
    assert len(lst) == 0


@pytest.mark.parametrize("version", torrents())
@pytest.mark.parametrize("progress", [0, 1, 2])
def test_mbtorrent(version, progress):
    """
    Test torrent creation for file size larger than 10MB.
    """
    tfile = tempfile(exp=26)
    outfile = str(tfile) + ".torrent"
    args = {
        "path": tfile,
        "progress": progress,
        "piece_length": "14",
        "outfile": outfile,
        "align": True,
    }
    torrent = version(**args)
    outfile, _ = torrent.write()
    assert os.path.exists(outfile)
    rmpath(tfile, outfile)


@pytest.mark.parametrize("params", [(10, 12), (10, 15), (20, 25), (28, 32)])
def test_progress_bar(params):
    """Testing the prog mixin with various sizes."""
    increment, total = params
    progbar = ProgMixin()
    progbar.prog_start(1 << total, "some/fake/path")
    while progbar.prog.state < total:
        progbar.prog_update(1 << increment)
    assert progbar.prog.state >= total



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Testing functions for the command line interface.
"""

import os
import sys
import datetime
import textwrap

import pyben
import pytest

from tests import dir1, dir2, file1, filemeta1, metafile1, rmpath, tempfile
from torrentfile import execute
from torrentfile.__main__ import main


def test_fix():
    """
    Test dir1 fixture is not None.
    """
    assert dir1 and dir2 and metafile1 and filemeta1 and file1


@pytest.fixture()
def folder(dir1):
    """
    Yield a folder object as fixture.
    """
    sfolder = str(dir1)
    torrent = sfolder + ".torrent"
    yield (sfolder, torrent)
    rmpath(torrent)


def test_cli_v1(folder):
    """
    Basic create torrent cli command.
    """
    folder, torrent = folder
    args = ["torrentfile", "create", folder, "-o", torrent]
    sys.argv = args
    execute()
    assert os.path.exists(torrent)


def test_cli_v2(folder):
    """
    Create torrent v2 cli command.
    """
    folder, torrent = folder
    args = [
        "torrentfile",
        "create",
        folder,
        "--meta-version",
        "2",
        "-o",
        torrent,
    ]
    sys.argv = args
    execute()
    assert os.path.exists(torrent)


def test_cli_v3(folder):
    """
    Create hybrid torrent cli command.
    """
    folder, torrent = folder
    args = [
        "torrentfile",
        "create",
        folder,
        "--meta-version",
        "3",
        "-o",
        torrent,
    ]
    sys.argv = args
    execute()
    assert os.path.exists(torrent)


def test_cli_private(folder):
    """
    Test private cli flag.
    """
    folder, torrent = folder
    args = ["torrentfile", "create", folder, "--private", "-o", torrent]
    sys.argv = args
    main()
    meta = pyben.load(torrent)
    assert "private" in meta["info"]


@pytest.mark.parametrize("piece_length", [2**exp for exp in range(14, 21)])
@pytest.mark.parametrize("version", ["1", "2", "3"])
def test_cli_piece_length(folder, piece_length, version):
    """
    Test piece length cli flag.
    """
    folder, torrent = folder
    args = [
        "torrentfile",
        "-v",
        "create",
        folder,
        "--piece-length",
        str(piece_length),
        "--meta-version",
        version,
        "--progress",
        "0",
        "-o",
        torrent,
    ]
    sys.argv = args
    execute()
    meta = pyben.load(torrent)
    assert meta["info"]["piece length"] == piece_length


@pytest.mark.parametrize("piece_length", [2**exp for exp in range(14, 21)])
@pytest.mark.parametrize("version", ["1", "2", "3"])
def test_cli_announce(folder, piece_length, version):
    """
    Test announce cli flag.
    """
    folder, torrent = folder
    args = [
        "torrentfile",
        "create",
        folder,
        "--piece-length",
        str(piece_length),
        "--meta-version",
        version,
        "--announce",
        "https://announce.org/tracker",
        "-o",
        torrent,
    ]
    sys.argv = args
    execute()
    meta = pyben.load(torrent)
    assert meta["announce"] == "https://announce.org/tracker"


@pytest.mark.parametrize("version", ["1", "2", "3"])
def test_cli_announce_list(folder, version):
    """
    Test announce-list cli flag.
    """
    folder, torrent = folder
    trackers = [
        "https://announce.org/tracker",
        "https://announce.net/tracker",
        "https://tracker.net/announce",
    ]
    args = [
        "torrentfile",
        "create",
        folder,
        "--meta-version",
        version,
        "-o",
        torrent,
        "--announce",
    ] + trackers
    sys.argv = args
    execute()
    meta = pyben.load(torrent)
    for url in trackers:
        assert url in [j for i in meta["announce-list"] for j in i]


@pytest.mark.parametrize("piece_length", [2**exp for exp in range(14, 21)])
@pytest.mark.parametrize("version", ["1", "2", "3"])
def test_cli_comment(folder, piece_length, version):
    """
    Test comment cli flag.
    """
    folder, torrent = folder
    args = [
        "torrentfile",
        "create",
        folder,
        "--piece-length",
        str(piece_length),
        "--meta-version",
        version,
        "--magnet",
        "--comment",
        "this is a comment",
        "--progress",
        "1",
        "-o",
        torrent,
    ]
    sys.argv = args
    execute()
    meta = pyben.load(torrent)
    assert meta["info"]["comment"] == "this is a comment"


@pytest.mark.parametrize("piece_length", [2**exp for exp in range(14, 21)])
@pytest.mark.parametrize("version", ["1", "2", "3"])
def test_cli_outfile(dir1, piece_length, version):
    """
    Test outfile cli flag.
    """
    outfile = dir1 + "test.torrent"
    args = [
        "torrentfile",
        "create",
        dir1,
        "--piece-length",
        str(piece_length),
        "--meta-version",
        version,
        "-o",
        outfile,
        "--prog",
        "1",
    ]
    sys.argv = args
    execute()
    assert os.path.exists(outfile)
    rmpath(outfile)


@pytest.mark.parametrize("piece_length", [2**exp for exp in range(14, 21)])
@pytest.mark.parametrize("version", ["1", "2", "3"])
def test_cli_creation_date(folder, piece_length, version):
    """
    Test if torrents created get an accurate timestamp.
    """
    folder, torrent = folder
    args = [
        "torrentfile",
        "create",
        folder,
        "--piece-length",
        str(piece_length),
        "--meta-version",
        version,
        "--comment",
        "this is a comment",
        "-o",
        torrent,
    ]
    sys.argv = args
    execute()
    meta = pyben.load(torrent)
    num = float(meta["creation date"])
    date = datetime.datetime.fromtimestamp(num)
    now = datetime.datetime.now()
    assert date.day == now.day
    assert date.year == now.year
    assert date.month == now.month


@pytest.mark.parametrize("piece_length", [2**exp for exp in range(14, 21)])
@pytest.mark.parametrize("version", ["1", "2", "3"])
def test_cli_created_by(folder, piece_length, version):
    """
    Test if created torrents recieve a created by field in meta info.
    """
    folder, torrent = folder
    args = [
        "torrentfile",
        "-q",
        "create",
        folder,
        "--piece-length",
        str(piece_length),
        "--meta-version",
        version,
        "--align",
        "--comment",
        "this is a comment",
        "-o",
        torrent,
    ]
    sys.argv = args
    execute()
    meta = pyben.load(torrent)
    assert "torrentfile" in meta["created by"]


@pytest.mark.parametrize("piece_length", [2**exp for exp in range(14, 21)])
@pytest.mark.parametrize("version", ["1", "2", "3"])
def test_cli_web_seeds(folder, piece_length, version):
    """
    Test if created torrents recieve a web seeds field in meta info.
    """
    folder, torrent = folder
    args = [
        "torrentfile",
        "create",
        folder,
        "--piece-length",
        str(piece_length),
        "--align",
        "--meta-version",
        version,
        "--web-seed",
        "https://webseed.url/1",
        "https://webseed.url/2",
        "https://webseed.url/3",
        "-o",
        torrent,
    ]
    sys.argv = args
    execute()
    meta = pyben.load(torrent)
    assert "https://webseed.url/1" in meta["url-list"]


@pytest.mark.parametrize("piece_length", [2**exp for exp in range(14, 21)])
@pytest.mark.parametrize("version", ["1", "2", "3"])
def test_cli_with_debug(folder, piece_length, version):
    """
    Test debug mode cli flag.
    """
    folder, torrent = folder
    args = [
        "torrentfile",
        "-v",
        "create",
        folder,
        "--piece-length",
        str(piece_length),
        "--meta-version",
        version,
        "--comment",
        "this is a comment",
        "-o",
        torrent,
    ]
    sys.argv = args
    execute()
    assert os.path.exists(torrent)


@pytest.mark.parametrize("piece_length", [2**exp for exp in range(14, 21)])
@pytest.mark.parametrize("version", ["1", "2", "3"])
def test_cli_with_source(folder, piece_length, version):
    """
    Test source cli flag.
    """
    folder, torrent = folder
    args = [
        "torrentfile",
        "create",
        folder,
        "--piece-length",
        str(piece_length),
        "--meta-version",
        version,
        "--source",
        "somesource",
        "-o",
        torrent,
    ]
    sys.argv = args
    execute()
    meta = pyben.load(torrent)
    assert meta["info"]["source"] == "somesource"


def test_cli_help():
    """
    Test showing help notice cli flag.
    """
    args = ["-h"]
    sys.argv = args
    try:
        assert execute()
    except SystemExit:
        assert True


@pytest.mark.parametrize("version", ["1", "2", "3"])
@pytest.mark.parametrize("progress", ["0", "1"])
def test_cli_empty_files(dir2, version, progress):
    """
    Test creating torrent with empty files.
    """
    outfile = str(dir2) + ".torrent"
    args = [
        "torrentfile",
        "create",
        str(dir2),
        "--meta-version",
        version,
        "--source",
        "somesource",
        "--prog",
        progress,
        "-o",
        outfile,
    ]
    sys.argv = args

    def walk(root, count):
        """
        Traverse directory to edit files.
        """
        if root.is_file():
            with open(root, "wb") as _:
                return 1
        elif root.is_dir():
            for item in root.iterdir():
                if count >= 2:
                    break
                count += walk(item, count)
        return count

    walk(dir2, 0)
    execute()
    assert os.path.exists(outfile)
    rmpath(outfile)


@pytest.mark.parametrize("ending", ["/", "\\"])
def test_cli_slash_path(dir1, ending):
    """
    Test if output when path ends with a /.
    """
    outfile = str(dir1) + ".torrent"
    if sys.platform != "win32" and ending == "\\":  # pragma: nocover
        ending = "/"
    args = [
        "torrentfile",
        "create",
        "-o",
        outfile,
        "-a",
        "https://announce1.org",
        "--private",
        str(dir1) + ending,
    ]
    sys.argv = args
    execute()
    assert os.path.exists(outfile)
    rmpath(outfile)


@pytest.mark.parametrize("sep", ["/", "\\"])
def test_cli_slash_outpath(dir1, sep):
    """
    Test if output when outpath ends with a /.
    """
    if sys.platform != "win32":
        sep = "/"  # pragma: nocover
    parent = os.path.dirname(dir1) + sep
    args = [
        "torrentfile",
        "create",
        "-a",
        "https://announce1.org",
        "--private",
        "-o",
        parent,
        str(dir1),
    ]
    sys.argv = args
    execute()
    outfile = str(dir1) + ".torrent"
    assert os.path.exists(outfile)
    rmpath(outfile)


@pytest.mark.parametrize("flag",
                         ["-a", "--announce", "--web-seed", "--http-seed"])
def test_cli_announce_path(dir1, flag):
    """
    Test CLI when path is placed after the trackers flag.
    """
    outfile = str(dir1) + ".torrent"
    args = [
        "torrentfile",
        "create",
        "-o",
        outfile,
        flag,
        "https://announce1.org",
        str(dir1),
    ]
    sys.argv = args
    execute()
    assert os.path.exists(outfile)
    rmpath(outfile)


@pytest.fixture()
def build(dir2):
    """Fixture for testing the build subcommand."""
    dest = os.path.join(os.path.dirname(__file__), "dest")
    if os.path.exists(dest):
        rmpath(dest)
        try:
            os.makedirs(dest)
        except FileExistsError:  # pragma: nocover
            rmpath(dest)
    return os.path.dirname(dir2), dest, dir2


@pytest.mark.parametrize("size", list(range(15, 19)))
@pytest.mark.parametrize("version", [1, 2])
def test_rebuild_subcommand(build, version, size):
    """Test the rebuild CLI subcommand."""
    basedir, dest, content = build
    args = [
        "torrentfile",
        "create",
        str(content),
        "--meta-version",
        str(version),
        "--piece-length",
        str(size),
        "-o",
        str(content) + ".torrent",
    ]
    sys.argv = args
    execute()
    args = ["torrentfile", "rebuild", "-m", basedir, "-c", basedir, "-d", dest]
    sys.argv = args
    counter = execute()
    assert counter > 0


def test_empty_maker(dir1):
    """Test empty create cli."""
    args = ["torrentfile", "create", dir1, "-o", dir1 + ".torrent"]
    sys.argv = args
    execute()
    assert os.path.exists(dir1 + ".torrent")
    rmpath(dir1 + ".torrent")


def test_rename():
    """Test the rename command."""
    tfile = str(tempfile())
    args = ["torrentfile", "create", tfile, "-o", tfile + ".torrent"]
    sys.argv = args
    execute()
    assert os.path.exists(tfile + ".torrent")
    parent = os.path.dirname(tfile)
    temp_path = os.path.join(parent, "renamed.torrent")
    os.rename(tfile + ".torrent", temp_path)
    print(os.listdir(parent))
    args = ["torrentfile", "rename", temp_path]
    sys.argv = args
    execute()
    assert os.path.exists(tfile + ".torrent")
    assert not os.path.exists(temp_path)
    rmpath(tfile + ".torrent", tfile)


@pytest.mark.parametrize("version", ["1", "2", "3"])
def test_cli_default_command(folder, version):
    """Test default command by ommitting command."""
    folder, torrent = folder
    args = [
        "torrentfile",
        "-q",
        folder,
        "--meta-version",
        version,
        "-o",
        torrent,
    ]
    sys.argv = args
    execute()
    assert os.path.exists(torrent)


def test_cli_configfile(folder):
    """Test config cli parameter."""
    folder, _ = folder
    args = ["torrentfile", "create", "--config", folder]
    sys.argv = args
    try:
        execute()
    except FileNotFoundError:
        assert True


@pytest.fixture
def configfile():
    """Test fixture for the configfile parameter."""
    config = """
    [config]
    meta-version = 3
    piece-length = 15
    announce = example
    private = true
    source = example
    comment = comment example
    """
    path = os.path.join(os.path.dirname(__file__), "torrentfile.ini")
    with open(path, "wt", encoding="utf8") as fd:
        fd.write(textwrap.dedent(config))
    yield path
    rmpath(path)


def test_cli_configfile_params(configfile, folder):
    """Testing parameters with config file option."""
    folder, torrent = folder
    args = [
        "torrentfile",
        "create",
        "--config",
        "--config-path",
        configfile,
        "-o",
        torrent,
        folder,
    ]
    sys.argv = args
    execute()
    meta = pyben.load(torrent)
    assert meta["info"]["private"] == 1
    assert meta["info"]["source"] == "example"
    assert meta["info"]["piece length"] == 2**15
    assert meta["announce"] == "example"
    assert meta["info"]["meta version"] == 2



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Testing functions for the progress module.
"""

import os
import sys
from pathlib import Path

from tests import (
    dir1, dir2, file1, file2, filemeta1, filemeta2, metafile1, metafile2,
    rmpath, sizedfiles, sizes)
from torrentfile.cli import main_script as main
from torrentfile.recheck import Checker
from torrentfile.utils import ArgumentError


def test_fixtures():
    """
    Test fixtures exist.
    """
    assert dir1 and dir2 and file1 and file2
    assert filemeta1 and filemeta2 and metafile1
    assert metafile2 and sizes and sizedfiles


def test_checker_class(dir1, metafile1):
    """
    Test Checker Class against meta files.
    """
    checker = Checker(metafile1, dir1)
    assert checker.results() == 100


def test_checker_first_piece(dir2, sizedfiles):
    """
    Test Checker Class when first piece is slightly alterred.
    """

    def change(path):
        """
        Change some bytes in file.
        """
        if path.is_file():
            new = b"Something other than what was there before."
            with open(path, "rb") as bfile:
                data = bfile.read()
            new_len = len(new)
            content = b"".join([new, data[new_len:]])
            with open(path, "wb") as bdoc:
                bdoc.write(content)
        elif path.is_dir():
            for item in path.iterdir():
                change(item)

    change(Path(dir2))
    checker = Checker(sizedfiles, dir2)
    assert checker.results() != 100


def test_checker_first_piece_alt(dir2, sizedfiles):
    """
    Test Checker Class when first piece is slightly alterred.
    """

    def change(path):
        """
        Change some bytes in file.
        """
        if os.path.isfile(path):
            with open(path, "rb") as bfile:
                data = bfile.read()
            new = b"some_other_bytes_to_use"
            new_len = len(new)
            with open(path, "wb") as wfile:
                wfile.write(new + data[new_len:])
        elif os.path.isdir(path):
            for item in os.listdir(path):
                change(os.path.join(path, item))

    change(dir2)
    checker = Checker(sizedfiles, dir2)
    assert checker.results() != 100


def test_partial_metafiles(dir2, sizedfiles):
    """
    Test Checker with data that is expected to be incomplete.
    """

    def shortenfile(path):
        """
        Shorten a few files for testing purposes.
        """
        with open(path, "rb") as bfile:
            data = bfile.read()
        with open(path, "wb") as bfile:
            bfile.write(data[:-(2**10)])

    for item in os.listdir(dir2):
        full = os.path.join(dir2, item)
        if os.path.isfile(full):
            shortenfile(full)

    testdir = os.path.dirname(dir2)
    checker = Checker(sizedfiles, testdir)
    assert checker.results() != 100


def test_checker_callback(dir1, metafile1):
    """
    Test Checker class with directory that points to nothing.
    """
    Checker.register_callback(lambda *x: print(x))
    checker = Checker(metafile1, str(dir1))
    assert checker.results() == 100


def test_checker_cli_args(dir1, metafile1):
    """
    Test exclusive Checker Mode CLI.
    """
    sys.argv = ["torrentfile", "check", str(metafile1), str(dir1)]
    output = main()
    assert output == 100


def test_checker_parent_dir(dir1, metafile1):
    """
    Test providing the parent directory for torrent checking feature.
    """
    checker = Checker(metafile1, os.path.dirname(dir1))
    assert checker.results() == 100


def test_checker_with_file(file1, filemeta1):
    """
    Test checker with single file torrent.
    """
    checker = Checker(filemeta1, file1)
    assert checker.results() == 100


def test_checker_no_meta_file():
    """
    Test Checker when incorrect metafile is provided.
    """
    try:
        Checker("peaches", "$")
    except FileNotFoundError:
        assert True


def test_checker_wrong_root_dir(metafile1):
    """
    Test Checker when incorrect root directory is provided.
    """
    try:
        Checker(metafile1, "fake")
    except FileNotFoundError:
        assert True


def test_checker_missing(sizedfiles, dir2):
    """
    Test Checker class when files are missing from contents.
    """
    count = 0
    for fd in Path(dir2).iterdir():
        if fd.is_file() and count < 2:
            rmpath(fd)
    checker = Checker(sizedfiles, dir2)
    assert int(checker.results()) < 100


def test_checker_class_allfiles(sizedfiles, dir2):
    """
    Test Checker class when all files are missing from contents.
    """

    def traverse(path):
        """
        Traverse internal subdirectories.
        """
        if path.is_file():
            rmpath(path)
        elif path.is_dir():
            for item in path.iterdir():
                traverse(item)

    traverse(dir2)
    checker = Checker(sizedfiles, dir2)
    assert int(checker.results()) < 100


def test_checker_class_allpaths(sizedfiles, dir2):
    """
    Test Checker class when all files are missing from contents.
    """
    for item in Path(str(dir2)).iterdir():
        rmpath(item)
    checker = Checker(sizedfiles, dir2)
    assert int(checker.results()) < 100


def test_checker_class_half_file(filemeta2, file2):
    """
    Test Checker class with half size single file.
    """
    half = int(os.path.getsize(file2) / 2)
    barr = bytearray(half)
    with open(file2, "rb") as content:
        content.readinto(barr)
    with open(file2, "wb") as content:
        content.write(barr)
    checker = Checker(filemeta2, file2)
    assert int(checker.results()) != 10


def test_checker_missing_singles(dir2, sizedfiles):
    """
    Test Checker class with half size single file.
    """

    def walk(root):
        """
        Remove first file found.
        """
        if root.is_file():
            rmpath(root)
            return True
        if root.is_dir():
            for item in root.iterdir():
                walk(item)
        return False

    walk(Path(dir2))
    checker = Checker(sizedfiles, dir2)
    assert int(checker.results()) < 100


def test_checker_result_property(dir1, metafile1):
    """
    Test Checker class with half size single file.
    """
    checker = Checker(metafile1, dir1)
    result = checker.results()
    assert checker.results() == result


def test_checker_simplest(dir1, metafile1):
    """
    Test the simplest example.
    """
    checker = Checker(metafile1, dir1)
    assert checker.results() == 100


def test_checker_empty_files(dir2, sizedfiles):
    """
    Test Checker when directory contains 0 length files.
    """

    def empty_files(root):
        """
        Dump contents of files.
        """
        if os.path.isfile(root):
            with open(root, "wb") as _:
                pass
            assert os.path.getsize(root) == 0
        elif os.path.isdir(root):
            for item in os.listdir(root):
                return empty_files(os.path.join(root, item))
        return root

    empty_files(dir2)
    checker = Checker(sizedfiles, dir2)
    assert checker.results() != 100


def test_recheck_wrong_dir(metafile1):
    """
    Test recheck function with directory that doesn't contain the contents.
    """
    grandparent = os.path.dirname(os.path.dirname(metafile1))
    try:
        _ = Checker(metafile1, grandparent)
    except FileNotFoundError:
        assert True


def test_recheck_mismatch_args(metafile1):
    """
    Test recheck function with mismatched directory.
    """
    grandparent = os.path.dirname(os.path.dirname(metafile1))
    try:
        _ = Checker(grandparent, metafile1)
    except ArgumentError:
        assert True



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Testing functions for the rebuild sub-action commands from command line args.
"""
import os

import pyben
import pytest

from tests import dir1, file1, file2, filemeta1, filemeta2, rmpath, tempfile
from torrentfile.commands import rebuild
from torrentfile.hasher import FileHasher, HasherHybrid, HasherV2
from torrentfile.rebuild import Assembler
from torrentfile.torrent import TorrentAssembler, TorrentFile


def test_fix():
    """Test that the fixtures import properly."""
    assert filemeta2 and dir1 and file1 and filemeta1 and file2


def create_torrentfile(path, creator, dest, piece_length):
    """
    Create a new torrent metafile based on the provided parameters.
    """
    torrent = creator(path=path, outfile=dest, piece_length=piece_length)
    data = torrent.write()
    return data


def create_dest():
    """
    Create the destination directory for testing the rebuild function.
    """
    parent = os.path.dirname(__file__)
    dest = os.path.join(parent, "dest2")
    if os.path.exists(dest):
        rmpath(dest)
        os.mkdir(dest)
    return dest


@pytest.mark.parametrize("creator", [TorrentFile, TorrentAssembler])
@pytest.mark.parametrize("size", list(range(15, 19)))
def test_rebuilder_with_dir(dir1, creator, size):
    """
    Test that the rebuilder works with full directories.
    """
    dest = create_dest()
    contents = [os.path.dirname(dir1)]
    outfile = dir1 + ".torrent"
    create_torrentfile(dir1, creator, outfile, size)
    assembler = Assembler(contents, contents, dest)
    counter = assembler.assemble_torrents()
    assert counter > 0


def get_params(path):
    """
    Shortcut method for building the parameters provided by fixtures.
    """
    base = os.path.dirname(path)
    out = os.path.join(os.path.dirname(base), "dest")
    if os.path.exists(out):
        rmpath(out)  # pragma: nocover
    os.mkdir(out)
    return ([path], [base], out)


@pytest.fixture()
def single1(filemeta1):
    """Test fixture for testing the build subcommand."""
    params = get_params(filemeta1)
    return params


@pytest.fixture()
def single2(filemeta2):
    """Test fixture testing build with mismatched file size."""
    params = get_params(filemeta2)
    return params


def test_single_file(single1):
    """
    Test functionality of single file torrent and single torrent.
    """
    assembler = Assembler(*single1)
    counter = assembler.assemble_torrents()
    assert counter > 0


def test_single_file_smaller(single2):
    """
    Test functionality of single file torrent and single torrent.
    """
    name = pyben.load(single2[0][0])["info"]["name"]
    contents = os.path.join(single2[1][0], name)
    with open(contents, "rb") as content:
        data = content.read()
    with open(contents, "wb") as content:
        content.write(data[:len(data) // 2])
    assembler = Assembler(*single2)
    counter = assembler.assemble_torrents()
    assert counter == 0


def test_wrong_path():
    """Test rebuild command with incorrect paths."""

    class Namespace:
        """
        Emulates the behaviour of argparse.Namespace.
        """

        metafiles = "/non/existent/path"
        destination = "/non/existing/path"
        contents = "/non/existing/path"

    try:
        rebuild(Namespace)
    except FileNotFoundError:
        assert True


@pytest.mark.parametrize("size", [2**i for i in range(14, 21)])
def test_file1_hashers(file2, size):
    """Test that all three of the version2 hashers produce the same result."""
    hasher1 = HasherHybrid(file2, size, progress=False)
    hasher2 = HasherV2(file2, size, progress=False)
    hasher3 = FileHasher(file2, size, progress=False)
    _ = list(hasher3)
    lst = [hasher1.root, hasher2.root, hasher3.root]
    print(lst)
    assert len(set(lst)) == 1


@pytest.fixture(scope="package")
def dextra():
    """Text fixture for creating testing directory."""
    paths = [
        "dir3/file1.bin",
        "dir3/subdir1/file2.xls",
        "dir3/subdir1/file3.mov",
        "dir3/subdir2/file4.txt",
        "dir3/subdir2/file5.md",
        "dir3/file6.png",
        "dir3/file7.nfo",
    ]
    temps, start = [], 13
    for path in paths:
        temp = tempfile(path=path, exp=start)
        start += 1
        temps.append(temp)
    return os.path.commonpath(temps)


@pytest.mark.parametrize("creator", [TorrentFile, TorrentAssembler])
@pytest.mark.parametrize("size", list(range(15, 19)))
def test_rebuild_extra_dir(dextra, creator, size):
    """Test rebuild with dir2 and an extra smaller file."""
    dest = create_dest()
    contents = [os.path.dirname(dextra)]
    outfile = str(dextra) + ".torrent"
    create_torrentfile(dextra, creator, outfile, size)
    assembler = Assembler(contents, contents, dest)
    counter = assembler.assemble_torrents()
    assert counter > 0



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Collection of classes that can be used as Mixins with other base classes.

Classes such as TorrentFile, TorrentFilev2, and all Hasher classes can use the
progress bar mixin.  And any class is eligible to use the callback mixin.
"""
import os
import sys
import math
import time
import shutil
from pathlib import Path

from torrentfile.utils import debug_is_on


class CbMixin:
    """
    Mixin class to set a callback hook during procedure.
    """

    @classmethod
    def cb(cls, *args, **kwargs):
        """Do nothing."""

    @classmethod
    def set_callback(cls, func):
        """
        Assign a callback to the Hashing class.

        Parameters
        ----------
        func : Callable
            the callback function
        """
        cls.cb = func  # pragma: nocover


class ProgressBar:
    """
    Holds the state and details of the terminal progress bars.

    Parameters
    ----------
    total : int
        the total amount to be accumulated.
    title : str
        the subject of the progress tracker
    length : int
        the width of the progress bar
    unit : str
        the text representation incremented
    start : int
        column where the progress bar should be drawn
    """

    def __init__(self, total: int, title: str, length: int, unit: str,
                 start: int):
        """
        Construct the progress bar object and store state of it's properties.
        """
        self.total = total
        self.start = start
        self.length = length
        self.fill = chr(9608)
        self.empty = chr(9617)
        self.state = 0
        self.unit = unit
        self.show_total = total
        if not unit:
            self.unit = ""  # pragma: nocover
        elif unit == "bytes":
            if self.total > 1_000_000_000:
                self.show_total = self.total / (2**30)
                self.unit = "GiB"
            elif self.total > 1_000_000:
                self.show_total = self.total / 1048576
                self.unit = "MiB"
            elif self.total > 10000:
                self.show_total = self.total / 1024
                self.unit = "KiB"
        self.suffix = f"/{self.show_total:.02f} {self.unit}"
        title = str(title)
        if len(title) > start:
            title = title[:start - 1]  # pragma: nocover
        padding = (start - len(title)) * " "
        self.prefix = "".join([title, padding])

    def get_progress(self) -> str:
        """
        Return the size of the filled portion of the progress bar.

        Returns
        -------
        str :
            the progress bar characters
        """
        if self.state >= self.total:
            fill = self.length
        else:
            fill = math.ceil((self.state / self.total) * self.length)
        empty = self.length - fill
        pbar = ["|", self.fill * fill, self.empty * empty, "| "]
        if self.unit == "GiB":
            state = self.state / (2**30)
        elif self.unit == "MiB":
            state = self.state / 1048576
        elif self.unit == "KiB":
            state = self.state / 1024
        else:
            state = self.state
        pbar.append(f"{state:.02f}")
        return "".join(pbar)


class ProgMixin:
    """
    Progress bar mixin class.

    Displays progress of hashing individual files, usefull when hashing
    really big files.
    """

    def prog_start(self,
                   total: int,
                   path: str,
                   length: int = 50,
                   unit: str = "bytes"):
        """
        Generate a new progress bar for the given file path.

        Parameters
        ----------
        total : int
            the total amount of units accumulating towards.
        path : str
            path to file being hashed.
        length : int
            the number of characters of the actual progress bar.
        unit : str
            the text representation of the value being measured.
        """
        title = path
        width = shutil.get_terminal_size().columns
        if len(str(title)) >= width // 2:
            parts = list(Path(title).parts)
            while (len("//".join(parts)) > (width // 2)) and (len(parts) > 0):
                del parts[0]
            if parts:
                title = os.path.join(*parts)
            else:
                title = os.path.basename(path)  # pragma: nocover
        length = min(length, width // 2)
        start = width - int(length * 1.5)
        self.prog = ProgressBar(total, title, length, unit, start)

    def prog_update(self, val: int):
        """
        Update progress bar.

        Using the value provided, increment the progress bar by that value.

        Parameters
        ----------
        val : int
            the number of bytes count the progress bar should increase.
        """
        if self.is_active():
            self.prog.state += val
            pbar = self.prog.get_progress()
            output = f"{self.prog.prefix}{pbar}{self.prog.suffix}\r"
            sys.stdout.write(output)
            sys.stdout.flush()

    def prog_close(self):
        """
        Finalize the last bits of progress bar.

        Increment the terminal by one line leaving the progress bar in place,
        and deleting the progress bar object to clear a space for the next one.
        """
        if self.is_active():
            sys.stdout.flush()
            sys.stdout.write("\n")
            del self.prog

    def is_active(self) -> bool:
        """
        Test to see if there is an active progress bar for object.

        Returns
        -------
        bool :
            True if there is, otherwise False.
        """
        if not debug_is_on() and hasattr(self, "prog"):
            return True
        return False


def waiting(msg: str, flag: list, timeout: int = 20):
    """
    Show loading message while thread completes processing.

    Parameters
    ----------
    msg : str
        Message string printed before the progress bar
    flag : list
        Once flag is filled exit loop
    timeout : int
        max amount of time to run the function.
    """
    then = time.time()
    codes, fill = list(range(9617, 9620)), chr(9619)
    size = idx = 0
    total = shutil.get_terminal_size().columns - len(msg) - 20

    def output(text: str):
        """
        Print parameter message to the console.

        Parameters
        ----------
        text : str
            output message
        """
        sys.stdout.write(text)
        sys.stdout.flush()

    output("\n")
    time.sleep(0.16)
    while len(flag) == 0:
        time.sleep(0.16)
        filled = (fill * size) + chr(codes[idx]) + (" " * (total - size))
        output(f"{msg}: {filled}\r")
        idx = idx + 1 if idx + 1 < len(codes) else 0
        size = size + 1 if size < total else 0
        if time.time() - then > timeout:
            break
    output("\n")



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Command Line Interface for TorrentFile project.

This module provides the primary command line argument parser for
the torrentfile package. The main_script function is automatically
invoked when called from command line, and parses accompanying arguments.

Functions
---------
main_script :
    process command line arguments and run program.
activate_logger :
    turns on debug mode and logging facility.

Classes
-------
Config : class
    controls logging configuration
TorrentFileHelpFormatter : HelpFormatter
    the command line help message formatter
"""

import io
import sys
import logging
from argparse import ArgumentParser, HelpFormatter

from torrentfile import commands
from torrentfile.utils import toggle_debug_mode
from torrentfile.version import __version__ as version


class Config:
    """
    Class the controls the logging configuration and output settings.

    Controls the logging level, or whether to app should operate in quiet mode.
    """

    @staticmethod
    def activate_quiet():
        """
        Activate quiet mode for the duration of the programs life.

        When quiet mode is enabled, no logging, progress or state information
        is output to the terminal
        """
        if sys.stdout or sys.stderr:
            sys.stdout = io.StringIO()
            sys.stderr = io.StringIO()

    @staticmethod
    def activate_logger():
        """
        Activate the builtin logging mechanism when passed debug flag from CLI.
        """
        logging.basicConfig(level=logging.WARNING)
        logger = logging.getLogger()
        console_handler = logging.StreamHandler(stream=sys.stderr)
        stream_formatter = logging.Formatter(
            "[%(asctime)s] [%(levelno)s] %(message)s",
            datefmt="%H:%M:%S",
            style="%",
        )
        console_handler.setFormatter(stream_formatter)
        console_handler.setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)
        logger.addHandler(console_handler)
        logger.debug("Debug: ON")
        toggle_debug_mode(True)


class TorrentFileHelpFormatter(HelpFormatter):
    """
    Formatting class for help tips provided by the CLI.

    Subclasses Argparse.HelpFormatter.
    """

    def __init__(self,
                 prog: str,
                 width: int = 45,
                 max_help_positions: int = 45):
        """
        Construct HelpFormat class for usage output.

        Parameters
        ----------
        prog : str
            Name of the program.
        width : int
            Max width of help message output.
        max_help_positions : int
            max length until line wrap.
        """
        super().__init__(prog,
                         width=width,
                         max_help_position=max_help_positions)

    def _split_lines(self, text: str, _: int) -> list:
        """
        Split multiline help messages and remove indentation.

        Parameters
        ----------
        text : str
            text that needs to be split
        _ : int
            max width for line.

        Returns
        -------
        list :
            split into multiline text list
        """
        lines = text.split("\n")
        return [line.strip() for line in lines if line]

    def _format_text(self, text: str) -> str:
        """
        Format text for cli usage messages.

        Parameters
        ----------
        text : str
            Pre-formatted text.

        Returns
        -------
        str
            Formatted text from input.
        """
        text = text % {"prog": self._prog} if "%(prog)" in text else text
        text = self._whitespace_matcher.sub(" ", text).strip()
        return text + "\n\n"

    def _join_parts(self, part_strings: list) -> str:
        """
        Combine different sections of the help message.

        Parameters
        ----------
        part_strings : list
            List of argument help messages and headers.

        Returns
        -------
        str
            Fully formatted help message for CLI.
        """
        parts = self._format_headers(part_strings)
        return super()._join_parts(parts)

    @staticmethod
    def _format_headers(parts: list) -> list:
        """
        Format help message section headers.

        Parameters
        ----------
        parts : list
            List of individual lines for help message.

        Returns
        -------
        list
            Input list with formatted section headers.
        """
        if parts and parts[0].startswith("usage:"):
            parts[0] = "Usage\n=====\n  " + parts[0][6:]
        headings = [i for i in range(len(parts)) if parts[i].endswith(":\n")]
        for i in headings[::-1]:
            parts[i] = parts[i][:-2].title()
            underline = "".join(["\n", "-" * len(parts[i]), "\n"])
            parts.insert(i + 1, underline)
        return parts


def execute(args: list = None) -> list:
    """
    Execute program with provided list of arguments.

    If no arguments are given then it defaults to using
    sys.argv.  This is the main entrypoint for the program
    and command line interface.

    Parameters
    ----------
    args : list
        Commandline arguments. default=None

    Returns
    -------
    list
        Depends on what the command line args were.
    """
    toggle_debug_mode(False)
    if not args:
        if sys.argv[1:]:
            args = sys.argv[1:]
        else:
            args = ["-h"]

    parser = ArgumentParser(
        "torrentfile",
        usage="torrentfile <options>",
        description=(
            "Command line tool for creating, editing, validating, building "
            "and interacting with all versions of Bittorrent files"),
        prefix_chars="-",
        formatter_class=TorrentFileHelpFormatter,
        conflict_handler="resolve",
    )

    parser.add_argument(
        "-q",
        "--quiet",
        help="Turn off all text output.",
        dest="quiet",
        action="store_true",
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"torrentfile v{version}",
        help="show program version and exit",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="debug",
        help="output debug information",
    )

    parser.set_defaults(func=parser.print_help)

    subparsers = parser.add_subparsers(
        title="Commands",
        dest="command",
        metavar="create, edit, info, magnet, recheck, rebuild, rename\n",
    )

    create_parser = subparsers.add_parser(
        "create",
        help="Create a new Bittorrent file.",
        prefix_chars="-",
        aliases=["new"],
        formatter_class=TorrentFileHelpFormatter,
    )

    create_parser.add_argument(
        "-a",
        "--announce",
        "--tracker",
        action="store",
        dest="announce",
        metavar="<url>",
        nargs="+",
        default=[],
        help="one or more space-seperated tracker url(s)",
    )

    create_parser.add_argument(
        "-p",
        "--private",
        action="store_true",
        dest="private",
        help="create private torrent",
    )

    create_parser.add_argument(
        "-s",
        "--source",
        action="store",
        dest="source",
        metavar="<source>",
        help="add source field to the metadata",
    )

    create_parser.add_argument(
        "--config",
        action="store_true",
        dest="config",
        help="""
        Parse torrent information from a config file. Looks in the current
        working directory, or the directory named .torrentfile in the users
        home directory for a torrentfile.ini file. See --config-path option.
        """,
    )

    create_parser.add_argument(
        "--config-path",
        action="store",
        metavar="<path>",
        dest="config_path",
        help="use in combination with --config to provide config file path",
    )

    create_parser.add_argument(
        "-m",
        "--magnet",
        action="store_true",
        dest="magnet",
    )

    create_parser.add_argument(
        "-c",
        "--comment",
        action="store",
        dest="comment",
        metavar="<comment>",
        help="include a comment in the torrent file metadata",
    )

    create_parser.add_argument(
        "-o",
        "--out",
        action="store",
        dest="outfile",
        metavar="<path>",
        help="path to write torrent file",
    )

    create_parser.add_argument(
        "--prog",
        "--progress",
        default="1",
        action="store",
        dest="progress",
        metavar="<int>",
        help="""
        set the progress bar level
        Options = 0, 1
        (0) = Do not display progress bar.
        (1) = Display progress bar.(default)
        """,
    )

    create_parser.add_argument(
        "--meta-version",
        default="1",
        choices=["1", "2", "3"],
        action="store",
        dest="meta_version",
        metavar="<int>",
        help="""
        bittorrent metafile version
        options = 1, 2, 3
        (1) = Bittorrent v1 (Default)
        (2) = Bittorrent v2
        (3) = Bittorrent v1 & v2 hybrid
        """,
    )

    create_parser.add_argument(
        "--piece-length",
        action="store",
        dest="piece_length",
        metavar="<int>",
        help="""
        (Default: auto calculated based on total size of content)
        acceptable values include numbers 14-26
        14 = 16KiB, 20 = 1MiB, 21 = 2MiB etc.  Examples:[--piece-length 14]
        """,
    )

    create_parser.add_argument(
        "--web-seed",
        action="store",
        dest="url_list",
        metavar="<url>",
        nargs="+",
        help="list of web addresses where torrent data exists (GetRight)",
    )

    create_parser.add_argument(
        "--http-seed",
        action="store",
        dest="httpseeds",
        metavar="<url>",
        nargs="+",
        help="list of URLs, addresses where content can be found (Hoffman)",
    )

    create_parser.add_argument(
        "--align",
        action="store_true",
        help=("Align pieces to file boundaries. "
              "This option is ignored when not used with V1 torrents."),
    )

    create_parser.add_argument(
        "content",
        action="store",
        metavar="<content>",
        nargs="?",
        help="path to content file or directory",
    )

    create_parser.set_defaults(func=commands.create)

    edit_parser = subparsers.add_parser(
        "edit",
        help="edit torrent file",
        prefix_chars="-",
        formatter_class=TorrentFileHelpFormatter,
    )

    edit_parser.add_argument(
        "metafile",
        action="store",
        help="path to *.torrent file",
        metavar="<*.torrent>",
    )

    edit_parser.add_argument(
        "--tracker",
        action="store",
        dest="announce",
        metavar="<url>",
        nargs="+",
        help="replace current trackers with one or more urls",
    )

    edit_parser.add_argument(
        "--web-seed",
        action="store",
        dest="url_list",
        metavar="<url>",
        nargs="+",
        help="replace current web-seed with one or more url(s)",
    )

    edit_parser.add_argument(
        "--http-seed",
        action="store",
        dest="httpseeds",
        metavar="<url>",
        nargs="+",
        help="replace current http-seed urls with new ones (Hoffman)",
    )

    edit_parser.add_argument(
        "--private",
        action="store_true",
        help="make torrent private",
        dest="private",
    )

    edit_parser.add_argument(
        "--comment",
        help="replaces any existing comment with <comment>",
        metavar="<comment>",
        dest="comment",
        action="store",
    )

    edit_parser.add_argument(
        "--source",
        action="store",
        dest="source",
        metavar="<source>",
        help="replaces current source with <source>",
    )

    edit_parser.set_defaults(func=commands.edit)

    info_parser = subparsers.add_parser(
        "info",
        help="show detailed information about a torrent file",
        prefix_chars="-",
        formatter_class=TorrentFileHelpFormatter,
    )

    info_parser.add_argument(
        "metafile",
        action="store",
        metavar="<*.torrent>",
        help="path to torrent file",
    )

    info_parser.set_defaults(func=commands.info)

    magnet_parser = subparsers.add_parser(
        "magnet",
        help="generate magnet url from an existing torrent file",
        aliases=["m"],
        prefix_chars="-",
        formatter_class=TorrentFileHelpFormatter,
    )

    magnet_parser.add_argument(
        "metafile",
        action="store",
        help="path to torrent file",
        metavar="<*.torrent>",
    )

    magnet_parser.add_argument(
        "--meta-version",
        action="store",
        choices=["0", "1", "2", "3"],
        default="0",
        help="""
        This option is only relevant for hybrid torrent files.
        Options = 0, 1, 2, 3
        (0) = [default] version is determined automatically
        (1) = create V1 magnet link only
        (2) = create V2 magnet link only
        (3) = create a hybrid magnet link
        """,
        dest="meta_version",
        metavar="<int>",
    )

    magnet_parser.set_defaults(func=commands.get_magnet)

    check_parser = subparsers.add_parser(
        "recheck",
        help="gives a detailed look at how much of the torrent is available",
        aliases=["check"],
        prefix_chars="-",
        formatter_class=TorrentFileHelpFormatter,
    )

    check_parser.add_argument(
        "metafile",
        action="store",
        metavar="<*.torrent>",
        help="path to .torrent file.",
    )

    check_parser.add_argument(
        "content",
        action="store",
        metavar="<content>",
        help="path to content file or directory",
    )

    check_parser.set_defaults(func=commands.recheck)

    rebuild_parser = subparsers.add_parser(
        "rebuild",
        help="""
        Re-assemble files obtained from a bittorrent file into the
        appropriate file structure for re-seeding.  Read documentation
        for more information, or use cases.
        """,
        formatter_class=TorrentFileHelpFormatter,
    )

    rebuild_parser.add_argument(
        "-m",
        "--metafiles",
        action="store",
        metavar="<*.torrent>",
        nargs="+",
        dest="metafiles",
        required=True,
        help="path(s) to .torrent file(s)/folder(s) containing .torrent files",
    )

    rebuild_parser.add_argument(
        "-c"
        "--contents",
        action="store",
        dest="contents",
        nargs="+",
        required=True,
        metavar="<contents>",
        help="folders that might contain the source contents needed to rebuld",
    )

    rebuild_parser.add_argument(
        "-d",
        "--destination",
        action="store",
        dest="destination",
        required=True,
        metavar="<destination>",
        help="path to where torrents will be re-assembled",
    )

    rebuild_parser.set_defaults(func=commands.rebuild)

    rename_parser = subparsers.add_parser(
        "rename",
        help="""Rename a torrent file to it's original name provided in the
                metadata/the same name you see in your torrent client.""",
        formatter_class=TorrentFileHelpFormatter,
    )

    rename_parser.add_argument(
        "target",
        action="store",
        metavar="<target>",
        help="path to torrent file",
    )

    rename_parser.set_defaults(func=commands.rename)

    all_commands = [
        "m",
        "-h",
        "-V",
        "new",
        "edit",
        "info",
        "check",
        "create",
        "magnet",
        "rename",
        "rebuild",
        "recheck",
    ]
    if not any(i for i in all_commands if i in args):
        start = 0
        while args[start] in ["-v", "-q"]:
            start += 1
        args.insert(start, "create")

    args = parser.parse_args(args)

    if args.quiet:
        Config.activate_quiet()

    elif args.debug:
        Config.activate_logger()

    if hasattr(args, "func"):
        return args.func(args)
    return args  # pragma: nocover


main_script = execute


def main() -> None:
    """
    Initiate main function for CLI script.
    """
    execute()



#! /usr/bin/python3
# -*- coding: utf-8 -*-

##############################################################################
#    Copyright (C) 2021-current alexpdev
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Testing the edit torrent feature.
"""

import sys

import pyben
import pytest

from tests import dir1, dir2, metafile2
from torrentfile.cli import main
from torrentfile.edit import edit_torrent


def test_fix():
    """
    Testing dir fixtures.
    """
    assert dir2 and metafile2 and dir1


@pytest.mark.parametrize(
    "announce", [["urla"], ["urlb", "urlc"], ["urla", "urlb", "urlc"]])
def test_edit_torrent(metafile2, announce):
    """
    Test edit torrent with announce param.
    """
    edits = {"announce": announce}
    data = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    assert data == meta
    assert data["announce-list"] == [announce]


@pytest.mark.parametrize("announce", ["urla", "urlb urlc", "urla urlb urlc"])
def test_edit_torrent_str(metafile2, announce):
    """
    Test edit torrent with announce param as string.
    """
    edits = {"announce": announce}
    data = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    assert data == meta
    assert data["announce-list"] == [announce.split()]


@pytest.mark.parametrize("url_list", ["urla", "urlb urlc", "urla urlb urlc"])
def test_edit_urllist_str(metafile2, url_list):
    """
    Test edit torrent with webseed param.
    """
    edits = {"url-list": url_list}
    data = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    assert data == meta
    assert data["url-list"] == url_list.split()


@pytest.mark.parametrize("httpseeds", ["urla", "urlb urlc", "urla urlb urlc"])
def test_edit_httpseeds_str(metafile2, httpseeds):
    """
    Test edit torrent with webseed param.
    """
    edits = {"httpseeds": httpseeds}
    data = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    assert data == meta
    assert data["httpseeds"] == httpseeds.split()


@pytest.mark.parametrize(
    "url_list", [["urla"], ["urlb", "urlc"], ["urla", "urlb", "urlc"]])
def test_edit_urllist(metafile2, url_list):
    """
    Test edit torrent with webseed param as string.
    """
    edits = {"url-list": url_list}
    data = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    assert data == meta
    assert data["url-list"] == url_list


@pytest.mark.parametrize(
    "httpseed", [["urla"], ["urlb", "urlc"], ["urla", "urlb", "urlc"]])
def test_edit_httpseeds(metafile2, httpseed):
    """
    Test edit torrent with webseed param as string.
    """
    edits = {"httpseeds": httpseed}
    data = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    assert data == meta
    assert data["httpseeds"] == httpseed


@pytest.mark.parametrize("comment", ["COMMENT", "COMIT", "MITCO"])
def test_edit_comment(metafile2, comment):
    """
    Test edit torrent with comment param.
    """
    edits = {"comment": comment}
    data = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    assert data == meta
    assert data["info"]["comment"] == comment


@pytest.mark.parametrize("source", ["SomeSource", "NoSouce", "MidSource"])
def test_edit_source(metafile2, source):
    """
    Test edit torrent with source param.
    """
    edits = {"source": source}
    data = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    assert data == meta
    assert data["info"]["source"] == source


def test_edit_private_true(metafile2):
    """
    Test edit torrent with private param.
    """
    edits = {"private": "1"}
    data = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    assert data == meta
    assert data["info"]["private"] == 1


def test_edit_private_false(metafile2):
    """
    Test edit torrent with private param False.
    """
    edits = {"private": ""}
    data = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    assert data == meta
    assert "private" not in data["info"]


def test_edit_none(metafile2):
    """
    Test edit torrent with None for all params.
    """
    edits = {
        "announce": None,
        "url-list": None,
        "comment": None,
        "source": None,
        "private": None,
    }
    data = pyben.load(metafile2)
    edited = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    assert data == meta == edited


def test_edit_removal(metafile2):
    """
    Test edit torrent with empty for all params.
    """
    edits = {
        "announce": "",
        "url-list": "",
        "httpseeds": "",
        "comment": "",
        "source": "",
        "private": "",
    }
    data = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    assert data == meta


@pytest.mark.parametrize("comment", ["commenta", "commentb", "commentc"])
@pytest.mark.parametrize("source", ["sourcea", "sourceb", "sourcec"])
@pytest.mark.parametrize("announce", [["url1", "url2", "url3"], ["url1"]])
@pytest.mark.parametrize("webseed", [["ftp1"], ["ftpa", "ftpb"]])
@pytest.mark.parametrize("httpseed", [["ftp1"], ["ftpa", "ftpb"]])
def test_edit_cli(metafile2, comment, source, announce, webseed, httpseed):
    """
    Test edit torrent with all params on cli.
    """
    sys.argv = [
        "torrentfile",
        "edit",
        metafile2,
        "--comment",
        comment,
        "--source",
        source,
        "--web-seed",
        webseed,
        "--http-seed",
        httpseed,
        "--tracker",
        announce,
        "--private",
    ]
    main()
    meta = pyben.load(metafile2)
    info = meta["info"]
    assert comment == info.get("comment")
    assert source == info.get("source")
    assert info.get("private") == 1
    assert meta["announce-list"] == [[announce]]
    assert meta["url-list"] == [webseed]


def test_metafile_edit_with_unicode(metafile2):
    """
    Test if editing full unicode works as it should.
    """
    edits = {
        "comment": "丂七万丈三与丏丑丒专且丕世丗両丢丣两严丩个丫丬中丮丯.torrent",
        "source": "丂七万丏丑严丩个丫丬中丮丯",
    }
    data = edit_torrent(metafile2, edits)
    meta = pyben.load(metafile2)
    com1 = data["info"]["comment"]
    com2 = meta["info"]["comment"]
    msg = edits["comment"]
    assert com1 == com2 == msg
